""" dmt/data/loading/oto_loader.py
OTO = OneToOne. Meaning 1 subject loads exactly 1 batch example. 

Daemon loader wraps torch's Dataloader by continuously loading into a queue.
'Daemon' here refers to its original software interpretation (not Python's)
where a process works continuously in the background.

Added Functionality:
  - Allows asynchronous loading from an additional single worker.
  - Adds headstart functionality so there's no need to wait for 1st batch & init
    - allows preloading of training set 
    - allows validation / test sets to load near end of training epoch
  - Wraps additional process information like memory & cpu usage
"""

import os
import psutil
import time
import logging, warnings
import multiprocessing
import torch


def batch_loading_worker(torch_loader, batch_queue):
    
    pid = os.getpid()
    this_process = psutil.Process(pid)
    
    start = time.time()
    for i, batch in enumerate(torch_loader):
        # print(f'Putting batch {i+1} in Q ({time.time() - start:.2f} sec).')
        batch_queue.put(batch)
        mem_usage = this_process.memory_info()[0]/2.**30
        logging.info(f'OTOWorker: Using {mem_usage:.2f} GB in PID {pid}')
        start = time.time()
    
    del torch_loader
    print('OTOLoader loading complete.')
    while True:  # wait for process to be killed
        time.sleep(1)


class OneToOneLoader:
    
    def __init__(
            self,
            *args, 
            headstart=False, 
            queue_maxsize=32,
            **kwargs):
        
        if multiprocessing.get_start_method() != 'fork':
            msg = ('With multiprocess dataloading, you should use the "fork" '
                   'start method for faster worker initialization times and ' 
                   'reduced memory usage (assuming you are only reading from '
                   'dataset components & not modifying them).')
            warnings.warn(msg)
        
        self.torch_loader = torch.utils.data.DataLoader(*args, **kwargs)
        self.batch_queue = torch.multiprocessing.Queue(maxsize=queue_maxsize)
        
        self.batch_loader = None
        self.loading = False
        self.batch_count = None
        self.iter_runs = 0
        
        if headstart:
            self.start_loading()
    
    @property
    def main_pid(self):
        return os.getpid()
    
    @property
    def worker_pid(self):
        if self.batch_loader is not None:
            return self.batch_loader.pid
        return None
    
    @property
    def worker_memory(self):
        if self.batch_loader is not None:
            process = psutil.Process(self.worker_pid)
            mem_usage_mb = process.memory_info().rss / (1024 ** 2)
            return mem_usage_mb
        return None
    
    def start_loading(self):
        if self.batch_loader != None and self.loading and self.iter_runs > 0:
            warnings.warn(f'(start_loading) Loader is already loading.')
            return
        
        logging.info('OTOLoader: Starting process to load batches')
        # print('OTOLoader: Starting process to load batches')
        args = (self.torch_loader, self.batch_queue)
        self.batch_loader = torch.multiprocessing.Process(
            target=batch_loading_worker, args=args)
        self.batch_loader.start()
        self.loading = True
    
    def stop_loading(self):
        self.loading = False
        if self.batch_loader is None:
            # warnings.warn(f'(stop_loading) Loader process not instantiated.')
            return
        if self.batch_loader.is_alive():
            logging.info('OTOLoader: Killing loader process to load batches.')
            # print('OTOLoader: Killing loader process to load batches.')
            self.batch_loader.kill()
            self.batch_loader.join()
            assert not self.batch_loader.is_alive()
        self.batch_loader = None
        logging.info('OTOLoader: Successfully killed loader process.')
        # print('OTOLoader: Successfully killed loader process.')
        self.iter_runs += 1
        
    def stop_loading_and_reset(self):
        """ Same as stop_loading except, it also clears the batch_queue. """
        start = time.time()
        self.stop_loading()
        
        logging.info('OTOLoader: Clearing batch queue.')
        from queue import Empty
        try:
            while True:  # clear queue
                self.batch_queue.get_nowait()
                print(f'OTOLoader deleted batch element. Qempty: '
                      f'{self.batch_queue.empty()}')  # this shouldn't print
        except Exception as e:
            if isinstance(e, Empty):
                logging.info('OTOLoader: Successfully cleared batch queue.')
            elif not self.batch_queue.empty():
                raise e
        elapsed = time.time() - start
        logging.info(f'OTOloader: took {elapsed:.2f} sec to kill loader.')
        return
    
    def __len__(self):
        return len(self.torch_loader)
    
    def __iter__(self):
        if not self.loading:
            self.stop_loading_and_reset()
            self.start_loading()
        self.batch_count = 0
        return self
    
    def __next__(self):
        if self.batch_count < len(self):
            batch = self.batch_queue.get()
            self.batch_count += 1
            return batch
        else:
            self.stop_loading_and_reset()
            raise StopIteration
        
    # def __del__(self):
    #     logging.info('OTOLoader: Deleting instance & components.')
    #     self.stop_loading_and_reset()
    #     del self.torch_loader
    #     self.torch_loader = None
    #     del self.batch_queue 
    #     self.batch_queue = None
        





import os
import numbers
import copy
from abc import ABC, abstractmethod
import torch

from dmt.data.samples.sample import Sample
from dmt.transforms.harmonizer import ImageHarmonizer
from dmt.utils.parse import parse_bool, parse_probability
from dmt.constants import (
    INTERPOLATIONS_2_SITK_INTERPOLATIONS as interpolations_d
)


class Transform(ABC):
    """ Abstract & base class for all image transforms.
    Implementation Structure:
        - all subclasses must define apply (opt. invertible) function
        - base class has parsing functions (input checking & processing)
    
    Transform inputs can be: 
        1. Raw Images: np.ndarray, torch.Tensor, PIL.Image, SimpleITK.Image
        2. Containers: rere.Image
    
    Args:
        p: Probability that this transform will be applied.
        copy: Make a shallow copy of the input before applying the transform.
        include_keys: Key names in samples that will be transformed if they
            are images.
        exclude_keys: Key names in samples that will be ignored. 
    """
    
    ### ------ #     Main API & Transformation Functionality      # ----- ###
    
    def __init__(self, p=1.0, copy=True, include_keys=None, exclude_keys=None):
        self.p = parse_probability(p)
        self.copy = parse_bool(copy)
        keys = Sample._parse_include_exclude_keys(include_keys, exclude_keys)
        self.include_keys, self.exclude_keys = keys
        
        self.transform_args = ()  # for reproducibility; updated in subclasses
        
    def __call__(self, data, include_keys=None, exclude_keys=None):
        """
        Handles functionality for all transforms:
            1. Determine if transform should be applied given probability.
            2. Standardize the image types given via a DataHarmonizer
            3. Copy sample if necessary.
            4. Transform the sample (rewrites new data to sample)
            5. Record the transformation arguments into new sample.
        Args:
            data: Can be a Sample, Image, Harmonizer, Torch tensor, Numpy array, 
                or SimpleITK Image.
        Returns:
            Transformed image(s) in the original type given. 
        """
        if torch.rand(1).item() > self.p:
            return data
        
        keys = include_keys, exclude_keys
        include_keys, exclude_keys = Sample._parse_include_exclude_keys(*keys)
        
        # overwrites keys specified in init if given here
        final_include_keys = self.include_keys
        if include_keys:  
            final_include_keys = include_keys 
        final_exclude_keys = self.exclude_keys
        if exclude_keys:
            final_exclude_keys = exclude_keys
        
        harmonizer = ImageHarmonizer(data)
        sample = harmonizer.get_sample()
        if self.copy:
            sample = copy.copy(sample)
            
        transformed_sample = self.apply_transform(sample)
        self._record_transform(transformed_sample)
        
        # convert to input type 
        out_data = harmonizer.get_output(transformed_sample)  
        return out_data
        
    @abstractmethod
    def apply_transform(self, sample):
        pass
    
    @property
    def name(self):
        return self.__class__.__name__
    
    @property
    def is_invertible(self):
        return hasattr(self, 'invert')
    
    ### ------ #      Transform Recording & Replication      # ----- ###
    
    def _record_transform(self, transformed_sample):
        reproducing_args = self.get_reproducing_arguments()
        transformed_sample.record_transform(reproducing_args)
    
    def get_reproducing_arguments(self):
        """
        Return a dictionary with the arguments that would be necessary to
        reproduce the transform exactly (arguments are from previous transform).
        """
        reproducing_arguments = {
            'name': self.name,
            'include_keys': self.include_keys,
            'exclude_keys': self.exclude_keys,
            'copy': self.copy,
            'p': self.p
        }
        t_args = {name: getattr(self, name) for name in self.transform_args}
        reproducing_arguments.update(t_args)
        return reproducing_arguments






import logging
import weakref

class Label(dict):
    """ Base class for all labels (e.g. masks, class ids). """
    
    # Class Variables
    _id_counter = 0  # global runtime ID generator for all Image subclasses
    _gid_to_instance = weakref.WeakValueDictionary()
    _reserved_attributes = ['num_classes', 'class_names']
    
    def __init__(self, **kwargs):
        
        # Global Label class ID tracking
        self.gid = Label._id_counter
        Label._gid_to_instance[self.gid] = self  # save memory
        Label._id_counter += 1
        
        super().__init__(**kwargs)  # update given attributes to dict
        self.update_attributes()  # make dot notation available
    
    @property
    def num_classes(self):
        raise NotImplementedError()  # weak enforcement
    
    @property
    def class_names(self):
        raise NotImplementedError()  # weak enforcement
    
    def get_one_hot(self):
        raise NotImplementedError()  # weak enforcement
    
    def get_ids(self):
        raise NotImplementedError()  # weak enforcement
    
    ### Functions below offers compatible/correct dict functionality
    ###   e.g. must update __dict__ after every update & override default funcs
    
    def __setattr__(self, name, value):
        """ Setting attrs in dot notation requires updating dict. """
        name = self._parse_new_attribute(name)
        self[name] = value
        super().__setattr__(name, value)
    
    def __getitem__(self, item):
        item = super().__getitem__(item)
        gid = None if not hasattr(self, 'gid') else self.gid
        logging.debug(f'Sample {gid}: Getting item: {item}')
        return item
    
    def __setitem__(self, key, value):
        key = self._parse_new_attribute(key)
        super().__setitem__(key, value)
        self.update_attributes()
        gid = None if not hasattr(self, 'gid') else self.gid
        logging.debug(f'Sample {gid}: Getting item: {value}')
        
    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                msg = f"Update expected at most 1 arguments, got {len(args)}."
                raise TypeError(msg)
            for k, v in dict(args[0]).items():
                k = self._parse_new_attribute(k)
                self[k] = v
        for k, v in kwargs.items():
            k = self._parse_new_attribute(k)
            self[k] = v
        self.update_attributes()
            
    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        self.update_attributes()
        return self[key]
    
    def update_attributes(self):
        # Allows attribute access through dot notation, e.g. image.spacing
        self.__dict__.update(self)
        
    def _parse_new_attribute(self, attribute):
        return attribute
        
        # TODO gatekeep user attribute setting
        msg = (f'Given attribute "{attribute}" is a reserved attribute '
               f'for class {type(self).__name__}.')
        assert attribute not in self._reserved_attributes, msg
        
        msg = f'Given attribute "{attribute}" cannot start with "_".'
        assert attribute[0] != '_', msg
        
        return attribute
    
    

    




import numpy as np



### ---- ### ---- \\    Getting Image Statistics     // ---- ### ---- ###

def print_image_summary(np_image):
    shape = np_image.shape
    shape_str = 'x'.join([str(s) for s in shape])
    dtype = str(np_image.dtype)
    max_val = np_image.max()
    max_str = f"{max_val}" if 'int' in str(type(max_val)) else f"{max_val:.2f}"
    min_val = np_image.min()
    min_str = f"{min_val}" if 'int' in str(type(min_val)) else f"{min_val:.2f}"
    print(f"{shape_str} {dtype} Image (max:{max_str} min:{min_str})")

def get_max_coord(np_image):
    r""" Gets the ndim coordinate for the max value in np_image. 
    Arguments:
        np_image (np.ndarray) - any dimensional image or array
    Returns:
        tuple of coordinate (only tuples exhibit proper array indexing behavior)
    """
    max_flat_index = np.argmax(np_image)
    coord = np.unravel_index(max_flat_index, np_image.shape)
    return coord


def get_min_coord(np_image):
    r""" Gets the ndim coordinate for the min value in np_image. 
    Arguments:
        np_image (np.ndarray) - any dimensional image or array
    Returns:
        tuple of coordinate (only tuples exhibit proper array indexing behavior)
    """
    max_flat_index = np.argmin(np_image)
    coord = np.unravel_index(max_flat_index, np_image.shape)
    return coord


def standardize_2dimage(standard_np_image):
    r""" Given a standard 2D numpy image, change image into a plottable form.
    A 'standard' numpy image is defined as:
     (1) Values are either {True, False} or [0, 1] or [0, 255]
     (2) Image is either gray (HxW, HxWx1, 1xHxW)
            or rgb (HxWx3, 3xHxW) or rgba (HxWx4, 4xHxW)
            
    Return:
        np_image that is [0, 255] uint8 & HxW or HxWx3 or HxWx4 based on type.
    """
    if not is_standard_2dimage(standard_np_image):
        print(f"(standardize_2dimage) WARNING: non-standard image given.")
        return standard_np_image
    
    np_type = str(standard_np_image.dtype)
    shape = standard_np_image.shape
    
    # convert to standard shape (HxW, HxWx[3,4])
    if len(shape) == 3:
        if shape[0] in (1, 3, 4):
            standard_np_image = to_channel_last(standard_np_image)
        new_shape = standard_np_image.shape
        if new_shape[-1] == 1:
            standard_np_image = standard_np_image.squeeze(-1)
    
    if 'bool' in np_type:
        return standard_np_image.astype(np.uint8) * 255
    if standard_np_image.max() <= 1:
        return (standard_np_image * 255).astype(np.uint8)
    return standard_np_image.astype(np.uint8)

    
def is_standard_2dimage(np_image):
    if np_image.ndim not in (2, 3):
        return False
    if 'bool' in str(np_image.dtype):
        return True
    if np_image.min() < 0 or np_image.max() > 255:  # separate line for optim
        return False
    shape = np_image.shape
    if np_image.ndim == 3:
        if shape[0] in (1, 3, 4) or shape[-1] in (1, 3, 4):
            return True
        return False
    return True


### ---- ### ---- \\    Image Basic Transforms     // ---- ### ---- ###

def resize(np_im, size, interp='bicubic'):
    pass

def gray_to_rgb(np_im, channel_last=True):
    if channel_last:
        return np.stack((np_im, np_im, np_im), axis=-1)
    return np.stack((np_im, np_im, np_im), axis=0)


def to_channel_last(np_im):
    return np.moveaxis(np_im, 0, -1)

def to_channel_first(np_im):
    return np.moveaxis(np_im, -1, 0)
    
    


""" Module data/samples/images.py 
Extensions of the base image class for generalized loading of most used
image formats in medical imaging research. 

Organization (class definitions in order): 
    VectorImage2D
    ScalarImage2D
    ScalarImage3D
"""

from __future__ import annotations  # allows Class annotations before definition
import sys, os
import pathlib
import logging

import torch
import PIL
import numpy as np
import SimpleITK as sitk

from ..utils.io import images2d as im2d_utils, images3d as im3d_utils
from ..utils.images import visualize as viz_utils
from .image_base import Image, VALID_INPUT_TYPES
from ..constants import (
    IMTYPE_2_SCALAR_SITK, 
    IMTYPE_2_VECTOR_SITK, 
    IMTYPE_2_NP, 
    IMTYPE_2_TORCH
)


### ---- ### ---- \\    2D Images     // ---- ### ---- ###

class VectorImage2D(Image):
    
    def __init__(
            self, 
            path_or_data,                             # image object
            container_type=sitk.sitkVectorUInt8,      # sitk container type
            permanent_load=False,
            **kwargs
            ):
        # Validate 
        self._validate_init_args(path_or_data, container_type)
        self._is_2d = True
        
        super().__init__(
            path_or_data, 
            container_type, 
            permanent_load=permanent_load, 
            **kwargs
        )
        logging.debug(f'🖼️ Image object created: {repr(self)}')
    
    def get_array(self, sitk_im, out_type, channel_first=False, **kwargs):
        msg = f'Given type "{out_type}" is not a valid numpy dtype.'
        assert out_type in IMTYPE_2_NP.values(), msg
        
        array = sitk.GetArrayFromImage(sitk_im)
        if array.dtype != out_type:
            array = array.astype(out_type)
        
        if channel_first:
            array = np.moveaxis(array, -1, 0)
        return array
    
    def plot(self, save=''):
        image = self.array
        title = self.name
        viz_utils.plot_2d_image(image, title=title, image_size=10, save=save)
    
    def _validate_init_args(self, path_or_data, container_type):
        msg = (f'Given type must be one of the following sitk vector types: '
               f'{IMTYPE_2_VECTOR_SITK.values()}')
        assert container_type in IMTYPE_2_VECTOR_SITK.values(), msg
        
        msg = (f'Input type (given: {type(path_or_data)}) is invalid. ' 
               f'Only these types are accepted: {VALID_INPUT_TYPES}')
        assert isinstance(path_or_data, VALID_INPUT_TYPES), msg
        
        if isinstance(path_or_data, (str, pathlib.Path)):
            msg = f"File doesn't exist: {path_or_data}"
            assert os.path.isfile(str(path_or_data)), msg
        elif isinstance(path_or_data, sitk.SimpleITK.Image):
            if path_or_data.GetPixelID() in IMTYPE_2_SCALAR_SITK.values():
                msg = f'If you input a sitk image, it must be a vector type.'
                raise ValueError(msg)
        elif isinstance(path_or_data, (np.ndarray, torch.Tensor)):
            shape = path_or_data.shape
            msg = f'Input array or tensor must have 3 relevant dimensions.'
            assert sum([i > 1 for i in shape]) == 3, msg
            msg = (f'Input image must be in channels-last format, and the last '
                   f'channel must have 3 dimensions (given: {shape}).')
            assert shape[-1] == 3, msg


class ScalarImage2D(Image):
    """
    Behavior
     - If a 2D RGB image is given in the form of data, then it will do its best
        to convert it to grayscale.
     - Images with any relevant dim of size 1 cannot be given as array or tensor
        e.g. 3x1 image <- dim 1 gonna be squeezed(), 4x4x1 or 1x4x4x1 OK
    """
    
    def __init__(
            self, 
            path_or_data,                       # image object
            container_type=sitk.sitkUInt8,      # sitk container type
            permanent_load=False,
            **kwargs
            ):
        # Validate 
        self._validate_init_args(path_or_data, container_type)
        path_or_data = self._process_image_input(path_or_data, container_type)
        self._is_2d = True
        
        super().__init__(
            path_or_data, 
            container_type, 
            permanent_load=permanent_load, 
            **kwargs
        )
        logging.debug(f'🖼️ Image object created: {repr(self)}')
    
    def get_array(self, sitk_im, out_type, **kwargs):
        msg = f'Given type "{out_type}" is not a valid numpy dtype.'
        assert out_type in IMTYPE_2_NP.values(), msg
        
        array = sitk.GetArrayFromImage(sitk_im)
        if array.dtype != out_type:
            array = array.astype(out_type)
        return array
    
    def plot(self, save=''):
        image = self.array
        title = self.name
        viz_utils.plot_2d_image(image, title=title, image_size=10, save=save)
    
    def _validate_init_args(self, path_or_data, container_type):
        msg = (f'Given type must be one of the following sitk vector types: '
               f'{IMTYPE_2_SCALAR_SITK.values()}')
        assert container_type in IMTYPE_2_SCALAR_SITK.values(), msg
        
        msg = (f'Input type (given: {type(path_or_data)}) is invalid. ' 
               f'Only these types are accepted: {VALID_INPUT_TYPES}')
        assert isinstance(path_or_data, VALID_INPUT_TYPES), msg
        
        if isinstance(path_or_data, (str, pathlib.Path)):
            msg = f"File doesn't exist: {path_or_data}"
            assert os.path.isfile(str(path_or_data)), msg    
        
    def _process_image_input(self, path_or_data, container_type):
        """ Turn any RGB image into a grayscale sitk if data is given. """
        if isinstance(path_or_data, (str, pathlib.Path)):
            path_or_data = str(path_or_data)
            return path_or_data
        
        # If sitk is a vector type object, convert to numpy then grayscale
        if isinstance(path_or_data, sitk.SimpleITK.Image):
            if path_or_data.GetPixelID() in IMTYPE_2_SCALAR_SITK.values():
                return path_or_data
            path_or_data = sitk.GetArrayFromImage(path_or_data)
        
        # If input is a tensor, convert it to numpy for further processing
        if isinstance(path_or_data, torch.Tensor):
            path_or_data = path_or_data.detach().cpu().numpy()
        
        # If numpy is given or has been processed, turn image into grayscale
        assert isinstance(path_or_data, np.ndarray), 'Oops, missed something'
        
        array = path_or_data.squeeze()
        if array.ndim == 2:  # gray image
            pass
        elif array.ndim == 3 and 3 in array.shape:  # assumed RGB
            assert array.shape[-1] == 3, 'Image must be in channel last format.'
            array = np.dot(array[...,:3], [0.2989, 0.5870, 0.1140])
            assert array.ndim == 2, 'Losing sanity rapidly..'
        else:
            msg = (f"Only grayscale images & RGB images can be inputted, "
               f"not image of shape {array.shape}.")
            raise ValueError(msg)

        sitk_im = sitk.GetImageFromArray(array, isVector=False)
        if sitk_im.GetPixelID() != container_type:
            sitk_im = sitk.Cast(sitk_im, container_type)
        return sitk_im
        
        
### ---- ### ---- \\    3D Images     // ---- ### ---- ###
        

class ScalarImage3D(Image):
    """ 
    Overloaded: get_array(), plot().
    Default Functionality: read_image(), get_tensor().
    """
    
    VALID_EXTS = ['.dcm', '.nii', '.nii.gz']
    
    def __init__(
            self, 
            path_or_data,                       # image object
            container_type=sitk.sitkInt16,      # sitk container type
            permanent_load=False,
            **kwargs
            ):
        # Validate 
        self._validate_init_args(path_or_data, container_type)
        self._is_2d = False
        
        super().__init__(
            path_or_data, 
            container_type, 
            permanent_load=permanent_load, 
            **kwargs
        )
        logging.debug(f'🖼️ Image object created: {repr(self)}')
    
    def get_array(self, sitk_im, out_type,
                  isotropic=False, interpolation='bspline', **kwargs):
        """
        Args:
            **kwargs: for compatibility and consistent API. 
        """
        msg = f'Given type "{out_type}" is not a valid numpy dtype.'
        assert out_type in IMTYPE_2_NP.values(), msg
        
        if isotropic:
            sitk_im = im3d_utils.resample_sitk_isotropic(
                            sitk_im, interpolation=interpolation, smooth=0)
        array = sitk.GetArrayFromImage(sitk_im)
        if array.dtype != out_type:
            array = array.astype(out_type)
        # print(self.type, np_image.dtype)
        return array
        
    def plot(self, save=''):
        image = self.array
        viz_utils.plot_3d_slices(image, save=save)
    
    def _validate_init_args(self, path_or_data, container_type):
        msg = (f'Given type must be one of the following sitk scalars: '
               f'{IMTYPE_2_SCALAR_SITK.values()}')
        assert container_type in IMTYPE_2_SCALAR_SITK.values(), msg
        
        msg = (f'Input type (given: {type(path_or_data)}) is invalid. ' 
               f'Only these types are accepted: {VALID_INPUT_TYPES}')
        assert isinstance(path_or_data, VALID_INPUT_TYPES), msg
        
        if isinstance(path_or_data, (str, pathlib.Path)):
            path_or_data = str(path_or_data)
            msg = f"File doesn't exist: {path_or_data}"
            assert os.path.isfile(str(path_or_data)), msg
            
            msg = f'Given file {path_or_data} must have {self.VALID_EXTS} exts.'
            cond = sum([path_or_data[-len(s):] == s for s in self.VALID_EXTS])
            assert cond, msg
        elif isinstance(path_or_data, sitk.SimpleITK.Image):
            if path_or_data.GetPixelID() in IMTYPE_2_VECTOR_SITK.values():
                msg = f'If you input a sitk image, it must be a scalar type.'
                raise ValueError(msg)
        elif isinstance(path_or_data, (np.ndarray, torch.Tensor)):
            shape = path_or_data.shape
            msg = f'Input array or tensor must have 3 relevant dimensions.'
            assert sum([i > 1 for i in shape]) == 3, msg
                





""" densenet.py (modified by: Charley Zhang 2020.08)
Modified implementation of PyTorch densenet from:
https://github.com/pytorch/vision/blob/master/torchvision/models/densenet.py

Model Overviews  (7 output classes)
-------------
DenseNet121: 6.9M params  | 74.43% Top1
DenseNet169: 12.5M params | 75.60% Top1
DenseNet201: 18.1M params | 76.90% Top1
DenseNet161: 26.5M params | 77.14% Top1
"""

import sys, os
import re
import torch
import torch.nn as nn
import torch.nn.functional as F
from collections import OrderedDict

from .basemodel import BaseModel

__all__ = ['DenseNet', 'densenet121', 'densenet169', 'densenet201', 'densenet161']


model_params_path = '/afs/crc.nd.edu/user/y/yzhang46/_DLResources/Models/'
model_params = {
    'densenet121': os.path.join(model_params_path, 'densenet121.pth'),
    'densenet169': os.path.join(model_params_path, 'densenet169.pth'),
    'densenet201': os.path.join(model_params_path, 'densenet201.pth'),
    'densenet161': os.path.join(model_params_path, 'densenet161.pth'),
}

def _get_params(model_name):
    assert model_name in model_params, f"{model_name} not in params list."
    return torch.load(model_params[model_name])


class ForwardHook:
    def __init__(self, module):
        self.activations = None
        self.forward_hook = module.register_forward_hook(self.forward_hook_fn)
        
    def forward_hook_fn(self, module, input, output):
        self.activations = output


class _DenseLayer(nn.Sequential):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate):
        super(_DenseLayer, self).__init__()
        self.add_module('norm1', nn.BatchNorm2d(num_input_features)),
        self.add_module('relu1', nn.ReLU(inplace=True)),
        self.add_module('conv1', nn.Conv2d(num_input_features, bn_size *
                        growth_rate, kernel_size=1, stride=1, bias=False)),
        self.add_module('norm2', nn.BatchNorm2d(bn_size * growth_rate)),
        self.add_module('relu2', nn.ReLU(inplace=True)),
        self.add_module('conv2', nn.Conv2d(bn_size * growth_rate, growth_rate,
                        kernel_size=3, stride=1, padding=1, bias=False)),
        self.drop_rate = drop_rate

    def forward(self, x):
        new_features = super(_DenseLayer, self).forward(x)
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, 
                                     training=self.training)
        return torch.cat([x, new_features], 1)


class _DenseBlock(nn.Sequential):
    def __init__(self, num_layers, num_input_features, bn_size, growth_rate, drop_rate):
        super(_DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = _DenseLayer(num_input_features + i * growth_rate, 
                                growth_rate, bn_size, drop_rate)
            self.add_module('denselayer%d' % (i + 1), layer)


class _Transition(nn.Sequential):
    def __init__(self, num_input_features, num_output_features):
        super(_Transition, self).__init__()
        self.add_module('norm', nn.BatchNorm2d(num_input_features))
        self.add_module('relu', nn.ReLU(inplace=True))
        self.add_module('conv', nn.Conv2d(num_input_features, num_output_features,
                                          kernel_size=1, stride=1, bias=False))
        self.add_module('pool', nn.AvgPool2d(kernel_size=2, stride=2))


class DenseNet(BaseModel):
    r"""Densenet-BC model class, based on
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_
    Args:
        growth_rate (int) - how many filters to add each layer (`k` in paper)
        block_config (list of 4 ints) - how many layers in each pooling block
        num_init_features (int) - the number of filters to learn in the first convolution layer
        bn_size (int) - multiplicative factor for number of bottle neck layers
          (i.e. bn_size * k features in the bottleneck layer)
        drop_rate (float) - dropout rate after each dense layer
        num_classes (int) - number of classification classes
    """

    def __init__(self, growth_rate=32, block_config=(6, 12, 24, 16),
                 num_init_features=64, bn_size=4, drop_rate=0.2, num_classes=1000,
                 final_drop_rate=0.):

        super(DenseNet, self).__init__()

        # First convolution
        self.features = nn.Sequential(OrderedDict([
            ('conv0', nn.Conv2d(3, num_init_features, kernel_size=7, stride=2, 
                                padding=3, bias=False)),
            ('norm0', nn.BatchNorm2d(num_init_features)),
            ('relu0', nn.ReLU(inplace=True)),
            ('pool0', nn.MaxPool2d(kernel_size=3, stride=2, padding=1)),
        ]))

        # Each denseblock
        self.hooks = []
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(num_layers=num_layers, num_input_features=num_features,
                                bn_size=bn_size, growth_rate=growth_rate, 
                                drop_rate=drop_rate)
            self.features.add_module('denseblock%d' % (i + 1), block)
            self.hooks.append(ForwardHook(block))
            
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = _Transition(num_input_features=num_features, 
                                    num_output_features=num_features // 2)
                self.features.add_module('transition%d' % (i + 1), trans)
                num_features = num_features // 2

        # Final batch norm
        self.features.add_module('norm5', nn.BatchNorm2d(num_features))
        self.final_drop_rate = final_drop_rate

        # Linear layer
        self.classifier = nn.Linear(num_features, num_classes)

        # Official init from torch repo.
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.constant_(m.bias, 0)

    def forward(self, x, activations=False, ret_dict=False):
        features = self.features(x)
        out = F.relu(features, inplace=True)
        out = F.avg_pool2d(out, kernel_size=7, stride=1).view(features.size(0), -1)
        if self.final_drop_rate > 0:
            # print('Final drop rate: ', self.final_drop_rate)
            out = F.dropout(out, p=self.final_drop_rate, 
                            training=self.training)
        activs = out
        out = self.classifier(out)
        
        if ret_dict:
            return {
                'features': features,
                'activations': activs,
                'out': out
            }
        
        if activations:  # backward compat.
            return out, activs
        else:
            return out


def densenet121(pretrained=False, **kwargs):
    r"""Densenet-121 model from
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    Structure:
        Bx64xH/4xW/4 @ 7x7+MP out
        Bx256xH/8xW/8 @ denseblock1 out (before transition //2 dims)
        Bx512xH/16xW/16 @ denseblock 2 out
        Bx1024xH/32xW/32 @ denseblock 3 out
        Bx1024xH/32xW/32 @ denseblock 4 out
    """
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 24, 16),
                     **kwargs)
    num_classes = kwargs['num_classes']
    if pretrained:
        # '.'s are no longer allowed in module names, but pervious _DenseLayer
        # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
        # They are also in the checkpoints in model_urls. This pattern is used
        # to find such keys.
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        state_dict = _get_params('densenet121')
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]
            if num_classes != 1000 and 'classifier' in key:
                del state_dict[key]
        load_state_dict(model, state_dict)
    return model


def densenet169(pretrained=False, **kwargs):
    r"""Densenet-169 model from
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    Structure:
        Bx64xH/4xW/4 @ 7x7+MP out
        Bx256xH/8xW/8 @ denseblock1 out (before transition //2 dims)
        Bx512xH/16xW/16 @ denseblock 2 out
        Bx1280xH/32xW/32 @ denseblock 3 out
        Bx1664xH/32xW/32 @ denseblock 4 out
    """
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 32, 32),
                     **kwargs)
    num_classes = kwargs['num_classes']
    if pretrained:
        # '.'s are no longer allowed in module names, but pervious _DenseLayer
        # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
        # They are also in the checkpoints in model_urls. This pattern is used
        # to find such keys.
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        state_dict = _get_params('densenet169')
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]
            if num_classes != 1000 and 'classifier' in key:
                del state_dict[key]
        load_state_dict(model, state_dict)
    return model


def densenet201(pretrained=False, **kwargs):
    r"""Densenet-201 model from
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    Structure:
        Bx64xH/4xW/4 @ 7x7+MP out
        Bx256xH/8xW/8 @ denseblock1 out (before transition //2 dims)
        Bx512xH/16xW/16 @ denseblock 2 out
        Bx1792xH/32xW/32 @ denseblock 3 out
        Bx1920xH/32xW/32 @ denseblock 4 out
    """
    model = DenseNet(num_init_features=64, growth_rate=32, block_config=(6, 12, 48, 32),
                     **kwargs)
    num_classes = kwargs['num_classes']
    if pretrained:
        # '.'s are no longer allowed in module names, but pervious _DenseLayer
        # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
        # They are also in the checkpoints in model_urls. This pattern is used
        # to find such keys.
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        state_dict = _get_params('densenet201')
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]
            if num_classes != 1000 and 'classifier' in key:
                del state_dict[key]
        load_state_dict(model, state_dict)
    return model


def densenet161(pretrained=False, **kwargs):
    r"""Densenet-161 model from
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`_
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    Structure:
        Bx96xH/4xW/4 @ 7x7+MP out
        Bx384xH/8xW/8 @ denseblock1 out (before transition //2 dims)
        Bx768xH/16xW/16 @ denseblock 2 out
        Bx2112xH/32xW/32 @ denseblock 3 out
        Bx2208xH/32xW/32 @ denseblock 4 out
    """
    model = DenseNet(num_init_features=96, growth_rate=48, block_config=(6, 12, 36, 24),
                     **kwargs)
    num_classes = kwargs['num_classes']
    if pretrained:
        # '.'s are no longer allowed in module names, but pervious _DenseLayer
        # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
        # They are also in the checkpoints in model_urls. This pattern is used
        # to find such keys.
        pattern = re.compile(
            r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
        state_dict = _get_params('densenet161')
        for key in list(state_dict.keys()):
            res = pattern.match(key)
            if res:
                new_key = res.group(1) + res.group(2)
                state_dict[new_key] = state_dict[key]
                del state_dict[key]
            if num_classes != 1000 and 'classifier' in key:
                del state_dict[key]
        load_state_dict(model, state_dict)
    return model


def load_state_dict(model, state_dict):
    print(' (DenseNet pretrained=True) Loading pretrained stat_dict..')
    print('\t', model.load_state_dict(state_dict, strict=False))


def get_model(name, pretrained=True, only_encoder=False, layer_drop_rate=0.2, 
              final_drop_rate=0., num_classes=1000):
    """
    Parameters (defaults are based on PyTorch values):
        pretrained (bool) - use ImageNet pretrained parameters
        only_encoder (bool) - only get feature extractor (full FMs)
        layer_drop_rate (probability) - drop rate after every dense layer
        final_drop_rate (probability) - drop rate before last pooling layer
        num_classes (int) - number of output classes (cut if only_encoder)
    """
    name = str(name)
    if '121' in name:
        model = densenet121(pretrained=pretrained, num_classes=num_classes,
                    drop_rate=layer_drop_rate, final_drop_rate=final_drop_rate)
    elif '169' in name:
        model = densenet169(pretrained=pretrained, num_classes=num_classes,
                    drop_rate=layer_drop_rate, final_drop_rate=final_drop_rate)
    elif '201' in name:
        model = densenet201(pretrained=pretrained, num_classes=num_classes,
                    drop_rate=layer_drop_rate, final_drop_rate=final_drop_rate)
    elif '161' in name:
        model = densenet161(pretrained=pretrained, num_classes=num_classes,
                    drop_rate=layer_drop_rate, final_drop_rate=final_drop_rate)
    else:
        raise ValueError(f"DenseNet name ({name}) is not supported.")
    
    if only_encoder:
        model = list(model.children())[0]

    print(f" (DenseNet get_model) {name} model (pretrained={pretrained})"
          f" loaded w/{model.param_counts[0]:,} params.")
    return model


""" Module: dmt/data/label_masks.py
Label subclasses that are all masks (usually for segmentation). 

TODO:
    - If I have a future project on VOC, add a VectorMask2D.
"""

import collections

import torch
import numpy as np
import SimpleITK as sitk

from .label_base import Label
from .images import ScalarImage3D, ScalarImage2D
from ..constants import IMTYPE_2_NP

from ..utils.io import images2d as im2d_utils, images3d as im3d_utils
from ..utils.images import visualize as viz_utils


class ScalarMask(Label):
    """ Base class for scalar masks. 
    API
        - To get raw mask data in np, tensor, or sitk form, use Image API
            e.g. mask.array, mask.tensor
        - get_one_hot_mask()
        - get_id_mask() 
    Basic mask input assumptions:
        - Masks only contain 1 channel & are non-negative
        - Each class has its own id & its id is mapped to a value via
            the 'class_values' argument.
    """
    
    def get_one_hot(self, crop=None, channel_first=False, tensor=False):
        """ Outputs an array or tensor that has a separate channel for each
        class. For pixels that belong to a class, value is 1, else 0. 
            NOTE: there's always gonna be at least 2 classes. (back, fore, etc)
        Args:
            channel_first: if True, output dimensions are CxHxW(xD)
            tensor: if True, returns uint8 tensor, else uint8 array.
        """
        if crop is None:
            id_mask = self.get_ids()
        else:
            id_mask = crop
        
        if channel_first:
            shape = [self.num_classes] + list(id_mask.shape)
            one_hot = np.zeros(shape, dtype=np.uint8)
            for c in range(self.num_classes):
                one_hot[c, ...][id_mask == c] = 1
        else:
            shape = list(id_mask.shape) + [self.num_classes]
            one_hot = np.zeros(shape, dtype=np.uint8)
            for c in range(self.num_classes):
                one_hot[..., c][id_mask == c] = 1
        
        if tensor:
            one_hot = torch.from_numpy(one_hot)  # takes numpy dtype, not float        
        
        return one_hot
    
    def get_ids(self, tensor=False):
        """ Outputs a mask that is the same shape as given mask, where each
        pixel has the value of its class ID. 
        Args:
            tensor: if True, returns uint32 tensor, else uint32 array.
        """
        array = self.array
        id_mask = np.zeros(array.shape, dtype=np.uint8)
        for c, val in enumerate(self.class_values):
            id_mask[array == val] = c
        
        if tensor:
            id_mask = torch.from_numpy(id_mask)  # takes numpy dtype, not float

        return id_mask
    
    ### --- Mask Attributes --- ###
    
    @property
    def class_names(self):
        return self._class_names
    
    @property
    def class_values(self):
        return self._class_values
    
    @property
    def num_classes(self):
        return self._num_classes
    
    @property
    def class_counts(self):
        if self._class_counts is None:
            self._update_mask_attributes()
        return self._class_counts
    
    @property
    def classes_distribution(self):
        counts = self.class_counts
        return counts / sum(counts)
    
    @property
    def unique_values(self):
        if self._unique_values is None:
            self._update_mask_attributes()
        return self._unique_values
    
    def _parse_class_variables(self, class_names, class_values):
        # Verify if given class_names is valid
        cn_type = type(class_names)
        msg = f'Argument "class_names" must be a sequence. Given: {cn_type}.'
        assert isinstance(class_names, collections.Sequence), msg
        
        try:
            class_names = [str(cn) for cn in list(class_names)]
        except:
            msg = ('Argument "class_names" must be a sequence where each '
                   'element can be represented as a string.')
            raise ValueError(msg)
        
        # Verify if class_values are correct
        if class_values is None or len(class_values) == 0:
            class_values = list(range(len(class_names)))  # init default values
        
        cv_type = type(class_values)
        msg = f'Argument "class_values" must be a sequence. Given: {cv_type}.'
        assert isinstance(class_values, collections.Sequence), msg

        try:
            class_values = [int(cv) for cv in list(class_values)]
        except:
            msg = ('Argument "class_values" must be a sequence where each '
                   'element can be represented as an integer.')
            raise ValueError(msg)
        
        msg = (f'Number of class_values ({len(class_values)} does not match '
               f'the number of classnames given ({len(class_names)})')
        assert len(class_names) == len(class_values), msg
            
        return class_names, class_values
    
    def _update_mask_attributes(self, sitk_im=None):
        """ Kept separate function in-case needed more specific func later.
        Args:
            sitk_im: given in __init__ so that sitk image does not have to be 
                reread from path if permanent_load is not on.
        """
        if sitk_im is None:
            sitk_im = self.sitk_image
        np_type = IMTYPE_2_NP[self.type]
        array = self.get_array(sitk_im, np_type, isotropic=False)
        unique, counts = np.unique(array, return_counts=True)
        
        num_classes = len(unique)
        msg = (f'Given {self.num_classes} classes, but this mask has '
               f'{num_classes} unique values!')
        assert len(self._class_names) >= num_classes, msg
        
        # check if unique values match up with associated class values
        given_vals = self._given_class_values
        if given_vals is None or len(given_vals) == 0:
            pre_msg = (f'During mask init, you gave no class_values, so it was '
                       f'defaulted to {self._class_values}. ')
        else:
            pre_msg = (f'During mask init, you gave these class_values: '
                       f'{self._class_values}.')
        post_msg = (f'When loading the image, it contained values '
                    f'({unique}) that were not defined in "class_values".')
        msg = pre_msg + post_msg
        for unique_val in unique:
            assert unique_val in self._class_values, msg
            
        class_counts = np.array([0] * self.num_classes)
        for u, c in zip(unique, counts):
            cid = self.class_values.index(u)
            class_counts[cid] = c
        self._class_counts = class_counts
        
        self._unique_values = unique


class ScalarMask2D(ScalarImage2D, ScalarMask):
    
    """
    Args:
        path_or_data: str, path, np, sitk, or tensor of mask. 
        class_names: list of class names with index = class-id
        class_values: list of intensities in mask image corresponding to class
            e.g. If in mask, 0=background, 100=kidney, 200=tumor, 
                then class_values would be [0, 100, 200] assuming that backgrnd,
                kidney, and tumor have class-ids of 0, 1, and 2, resp. 
            The default class_values are set to range(C), C = # of classes
        container_type: storage type of mask (defaults to uint8)
        permanent_load: flag for mask data to remain in RAM or not
        **kwargs: additional attributes of the mask to be stored. 
    """
    
    def __init__(
            self, 
            path_or_data,                      # image object (most often path)
            class_names,                       # list of classnames
            class_values=None,
            container_type=sitk.sitkUInt8,     # container type
            permanent_load=True,               # low comp overhead for masks
            **kwargs
            ):
        
        super().__init__(   # verifies image stuff
            path_or_data, 
            container_type, 
            permanent_load=permanent_load, 
            **kwargs)
        
        self._given_class_values = class_values  # for informative error msg
        class_names, class_values = self._parse_class_variables(class_names, 
                                                                class_values)
        
        # Store attributes
        self._class_names = class_names
        self._class_values = class_values
        self._num_classes = len(class_names)
        
        self._class_counts = None
        self._unique_values = None
        if self.permanent_load:
            self._update_mask_attributes(sitk_im=self.sitk_image)
    
    
    ### ---- ### ---- \\    Mask Label Properties     // ---- ### ---- ###
        
    def get_array(self, sitk_im, out_type,
                  interpolation='nearest', **kwargs):
        msg = f'Given type "{out_type}" is not a valid numpy dtype.'
        assert out_type in IMTYPE_2_NP.values(), msg
        
        array = sitk.GetArrayFromImage(sitk_im)
        if array.dtype != out_type:
            array = array.astype(out_type)
        return array


class ScalarMask3D(ScalarImage3D, ScalarMask):
    """
    Args:
        path_or_data: str, path, np, sitk, or tensor of mask. 
        class_names: list of class names with index = class-id
        class_values: list of intensities in mask image corresponding to class
            e.g. If in mask, 0=background, 100=kidney, 200=tumor, 
                then class_values would be [0, 100, 200] assuming that backgrnd,
                kidney, and tumor have class-ids of 0, 1, and 2, resp. 
            The default class_values are set to range(C), C = # of classes
        container_type: storage type of mask (defaults to uint8)
        permanent_load: flag for mask data to remain in RAM or not
        **kwargs: additional attributes of the mask to be stored. 
    """
    
    def __init__(
            self, 
            path_or_data,                      # image object (most often path)
            class_names,                       # list of classnames
            class_values=None,
            container_type=sitk.sitkUInt8,     # container type
            permanent_load=True,               # low comp overhead for masks
            **kwargs
            ):
        
        super().__init__(   # verifies image stuff
            path_or_data, 
            container_type, 
            permanent_load=permanent_load, 
            **kwargs)
        
        self._given_class_values = class_values  # for informative error msg
        class_names, class_values = self._parse_class_variables(class_names, 
                                                                class_values)
        
        # Store attributes
        self._class_names = class_names
        self._class_values = class_values
        self._num_classes = len(class_names)
        
        self._class_counts = None
        self._unique_values = None
        if self.permanent_load:
            self._update_mask_attributes(sitk_im=self.sitk_image)
    
    ### ---- ### ---- \\    Mask Label Properties     // ---- ### ---- ###
    
    def get_array(self, sitk_im, out_type,
                  isotropic=False, interpolation='nearest', **kwargs):
        """ OVERLOAD - if you want custom sitk to numpy conversion in subclass. 
        This function overloads the get_array() from Image. 
        """
        msg = f'Given type "{out_type}" is not a valid numpy dtype.'
        assert out_type in IMTYPE_2_NP.values(), msg
        
        if isotropic:
            sitk_im = im3d_utils.resample_sitk_isotropic(
                            sitk_im, interpolation=interpolation, smooth=0)
        array = sitk.GetArrayFromImage(sitk_im)
        if array.dtype != out_type:
            array = array.astype(out_type)
        # print(self.type, np_image.dtype)
        return array
    
    
    
    
        
    
    

                



""" Module dmt/utils/queue.py

Implements custom multi-processing queue that is faster for big data transfer
and avoids the problem of slow buffer speeds in Q.get() in multiprocessing.
"""

import logging
import random
import multiprocessing
import torch
import queue
import weakref
import threading
from threading import Thread


class Counter:
    """ Process-safe counter. """
    def __init__(self, start_value=0):
        self._count = multiprocessing.Value('i', 0)
        
    @property
    def count(self):
        return self._count.value
        
    def increment(self, n=1):
        with self._count.get_lock():
            self._count.value += n
    
    def decrement(self, n=1):
        with self._count.get_lock():
            self._count.value -= n


class RandomQueue:
    """ A process-safe Queue that returns a random element within
        the queue. This is done via random priorities give in put().
    Components:
        - torch.multiprocessing.Queue
        - queue.PriorityQueue
        - daemon that uses a separate thread to load from Q to PQ
    """
    
    def __init__(self, maxsize=0, shuffle=True):
        import _multiprocessing
        python_maxsize = _multiprocessing.SemLock.SEM_VALUE_MAX
        if maxsize <= 0:
            maxsize = python_maxsize
        self._maxsize = min(maxsize, python_maxsize // 2)
        
        self._queue = torch.multiprocessing.Queue(maxsize=self._maxsize)
        self._pqueue = torch.multiprocessing.Queue(maxsize=self._maxsize)
        # self._pqueue = queue.PriorityQueue(maxsize=self._maxsize)
        self._size = Counter(start_value=0)
        
        self.kill_event = threading.Event()
        self.shuffle = shuffle
        self.loading = False
        self.transfer_thread = None
        self.start_loading()
    
    def put(self, item, block=True, timeout=None):
        self._queue.put(item, block=block, timeout=timeout)
        self._size.increment()
        
    def get(self, block=True, timeout=None):
        item = self._pqueue.get(block=block, timeout=timeout)
        self._size.decrement()
        return item[1]
    
    def get_nowait(self):
        item = self._pqueue.get_nowait()
        self._size.decrement()
        return item[1]

    def qsize(self):
        return self._size.count
    
    def empty(self):
        return True if self._size.count == 0 else False

    def full(self):
        return True if self._size.count == self._maxsize else False
    
    def start_loading(self):
        if self.loading:
            return
        args = self._queue, self._pqueue, self, self.kill_event
        self.transfer_thread = RandomQueue._start_transfer_daemon(*args)
        self.loading = True
        
    def delete(self):
        self._queue.put('kill_me')
        clear_queue(self._pqueue)
        while self.transfer_thread.is_alive():
            time.sleep(0.25)
        assert not self.transfer_thread.is_alive()
        self.transfer_thread = None
    
    @staticmethod
    def _start_transfer_daemon(src_q, dst_q, me, kill_event):
        
        def transfer(src_q, dst_q, kill_event):
            wait_flag = True
            transfer_items = set()
            while not kill_event.is_set():
                try:
                    obj = src_q.get(block=False, timeout=0.1)
                except:
                    obj = None
                # if isinstance(obj, str) and obj == 'kill_me':
                #     print(f'(FastQ-daemon) Got Kill. Quitting daemon thread.')
                #     break
                if obj is not None:
                    transfer_items.add(obj)
                    
                if not transfer_items or (wait_flag and len(transfer_items) < 24):
                    continue
                else:
                    wait_flag = False
                    priority = random.randint(1, 1000000)
                    item = transfer_items.pop()
                    priority_item = (priority, item)
                    dst_q.put(priority_item)
            print(f'Transfer thread daemon killed.')
            logging.info('RandomQueue\'s transfer daemon process is exiting')
        
        def stop_daemon(ref):
            print(f'(FastQ) Stop thread initiated')
            # src_q.put('kill_me')
            kill_event.set()
        
        me_weak = weakref.ref(me, stop_daemon)
        args = (src_q, dst_q, kill_event)
        transfer_thread = Thread(target=transfer, args=args)
        transfer_thread.daemon = True
        transfer_thread.start()
        return transfer_thread
    
    
        
    
        

"""
class FastQueue:
    # Wraps mp.Queue.get() with a thread daemon that moves it immediately
    # to a threaded queue so no buffering is necessary when get is called.
    
    def __init__(self, maxsize=0):
        if maxsize <= 0:
            import _multiprocessing
            maxsize = _multiprocessing.SemLock.SEM_VALUE_MAX
        self.maxsize = maxsize
        
        self.mpq = torch.multiprocessing.Queue(maxsize=maxsize)
        self.qq = queue.Queue(maxsize=maxsize)

        FastQueue._start_transfer_daemon(self.mpq, self.qq, self)

    def __del__(self):
        self.mpg.close()
        del self.mpq
        self.qq.close()
        del self.qq

    def put(self, item, block=True, timeout=None):
        self.mpq.put(item, block=block, timeout=timeout)

    def get(self, block=True, timeout=None):
        return self.qq.get(block=block, timeout=timeout)

    def qsize(self):
        return self.qq.qsize() + self.mpq.qsize()

    def empty(self):
        return self.qq.empty() and self.mpq.empty()

    def full(self):
        return self.qq.full() and self.mpq.full()

    @staticmethod
    def _start_transfer_daemon(src_q, dst_q, me):
        sentinel = object()

        def transfer(src_q, dst_q, me_ref):
            while me_ref():
                obj = src_q.get(block=True)
                if obj is sentinel:
                    print(f'(FastQ-daemon) End reached. Quitting daemon thread.')
                    break
                dst_q.put(obj)
                    
                # print 'steal'
            # print 'daemon done'
        
        def stop_daemon(ref):
            print(f'(FastQ) Stop thread initiated')
            src_q.put(sentinel)

        me1 = weakref.ref(me, stop_daemon)
        transfer_thread = Thread(target=transfer, args=(src_q, dst_q, me1,))
        transfer_thread.daemon = True
        transfer_thread.start()
"""

def clear_queue(queue):
    from queue import Empty
    try:
        while True:
            queue.get_nowait()
    except Empty:
        return


if __name__ == '__main__':
    print('hi')
    rq = RandomQueue()
    print(f'RandomQueue created (maxsize = {rq._maxsize})')
    
    import time
    start = time.time()
    rq.put(torch.rand((128, 128, 64)))
    rq.put(torch.rand((128, 128, 65)))
    rq.put(torch.rand((128, 128, 66)))
    rq.put(torch.rand((128, 128, 67)))
    rq.put(torch.rand((128, 128, 68)))
    rq.put(torch.rand((512, 512, 69)))
    rq.put(torch.rand((512, 512, 70)))
    time.sleep(2)
    
    while not rq.empty():
        print(rq.get().shape)
    print(f'Took {time.time() - start:2f} sec')
    rq.delete()
        


""" EfficientNet Implementation from PyPi timm


EfficientNet Overview (from: https://arxiv.org/pdf/1905.11946v5.pdf)
---------------------
EfficientNet-B0 - 77.1% top1, 5.3M params, 0.39B FLOPs
EfficientNet-B1 - 79.1% top1, 7.8M params, 0.70B FLOPS
EfficientNet-B2 - 80.1% top1, 9.2M params, 1.0B  FLOPs
EfficientNet-B3 - 81.6% top1, 12M  params, 1.8B  FLOPs
EfficientNet-B4 - 82.9% top1, 19M  params, 4.2B  FLOPs
EfficientNet-B5 - 83.6% top1, 30M  params, 9.9B  FLOPs
EfficientNet-B6 - 84.0% top1, 43M  params, 19B   FLOPs
EfficientNet-B7 - 84.3% top1, 66M  params, 37B   FLOPs

Resources
----------
Pretrained Weights / Model Implementations:
    https://github.com/rwightman/pytorch-image-models/

"""

import os
import timm

import torch
import torch.nn as nn
import torch.nn.functional as F


from .basemodel import BaseModel


model_params_path = '/afs/crc.nd.edu/user/y/yzhang46/_DLResources/Models/'
model_params = {
    'b0': os.path.join(model_params_path, 'efficientnetb0.pth'),
    'b1': os.path.join(model_params_path, 'efficientnetb1.pth'),
    'b2': os.path.join(model_params_path, 'efficientnetb2.pth'),
    'b3': os.path.join(model_params_path, 'efficientnetb3.pth'),
}


def load_state_dict(model, state_dict, load_classifier_params=False):
    if not load_classifier_params:
        del_keys = [k for k in state_dict.keys() if "classifier" in k]
        for k in del_keys:
            state_dict.pop(k, None)
    print(' (EfficientNet pretrained=True) Loading pretrained stat_dict..')
    print('\t', model.load_state_dict(state_dict, strict=False))

def get_model(name, pretrained=True, num_classes=1000, only_encoder=False):
    """
    Parameters (defaults are based on PyTorch values):
        pretrained (bool) - use ImageNet pretrained parameters
        only_encoder (bool) - only get feature extractor (full FMs)
        layer_drop_rate (probability) - drop rate after every dense layer
        final_drop_rate (probability) - drop rate before last pooling layer
        num_classes (int) - number of output classes (cut if only_encoder)
    """
    name = str(name)
    if 'b0' in name:
        model = timm.create_model('efficientnet_b0', pretrained=False,
        	num_classes=num_classes)
       	params_key = 'b0'
    elif 'b1' in name:
        model = timm.create_model('efficientnet_b1', pretrained=False,
            num_classes=num_classes)
        params_key = 'b1'
    elif 'b2' in name:
        model = timm.create_model('efficientnet_b2', pretrained=False,
            num_classes=num_classes)
        params_key = 'b2'
    elif 'b3' in name:
        model = timm.create_model('efficientnet_b3', pretrained=False,
            num_classes=num_classes)
        params_key = 'b3'
    else:
        raise ValueError(f"DenseNet name ({name}) is not supported.")

    if pretrained:
        cdiff = False if not pretrained or num_classes != 1000 else True
        state_dict = torch.load(model_params[params_key])
        load_state_dict(model, state_dict, load_classifier_params=cdiff)
    
    if only_encoder:
        raise NotImplementedError()
        model = list(model.children())[0]

    N_params = sum([p.numel() for p in model.parameters()])
    print(f" (EfficientNet get_model) {name} model (pretrained={pretrained})"
          f" loaded w/{N_params:,} params.")
    return model


""" bigunet2d.py
Larger U-Net PyTorch implementation from:
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..base_model import BaseModel


def get_model(in_channels=1, out_channels=1, bilinear_up=True, base_size=80):
    """
    Args:
        in_channels: # of input image channels.
        out_channels: # of output classes (specifically output dims).
        bilinear_up: Flag to use bilinear interpolation or upconvolution.
        base_size: Number of channels for base convolution (64 on std unet).
    """
    model = BigUNet2D(in_channels, out_channels, 
                      bilinear=bilinear_up, base_size=base_size)
    return model



### ======================================================================== ###
### * ### * ### * ### *          U-Net Components        * ### * ### * ### * ###
### ======================================================================== ###


class DoubleConv(nn.Module):
    """ (Same-Conv => [BN] => ReLU) * 2 """
    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)


# class BottleNeck(nn.Module):
#     def __init__(self, in_channels, out_channels)


class Down(nn.Module):
    """ 2x Downscale MP -> Double Conv """
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """ 2x Upscale -> Double Conv """

    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        # if bilinear, use normal convolutions to reduce the number of channels
        if bilinear:
            self.up = nn.Upsample(
                scale_factor=2, 
                mode='bilinear', 
                align_corners=True
            )
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(
                in_channels , 
                in_channels // 2, 
                kernel_size=2, 
                stride=2
            )
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    """ 1x1 Conv for Last Layer """
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)


### ======================================================================== ###
### * ### * ### * ### *               U-Net              * ### * ### * ### * ###
### ======================================================================== ###


class BigUNet2D(BaseModel):
    def __init__(self, n_channels, n_classes, bilinear=True, base_size=80):
        """
        Parameters
            n_channels - number of input channels in image
            n_classes - number of output channels that model will output
            bilinear (bool) - use bilinear upsampling or transposed conv
            base_size (int) - std U-Net: 64, 80 is 25% inc
        """
        super(BigUNet2D, self).__init__()
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.bilinear = bilinear
        self.base_size = base_size
        if self.base_size % 2 == 1:
            self.base_size -= 1  # make channels even
        print(f"  Prepping BigUNet w/base_size {self.base_size}.")

        self.inc = DoubleConv(n_channels, self.base_size)
        self.down1 = Down(self.base_size, self.base_size * 2)
        self.down2 = Down(self.base_size * 2, self.base_size * 4)
        self.down3 = Down(self.base_size * 4, self.base_size * 8)
        factor = 2 if bilinear else 1
        self.down4 = Down(self.base_size * 8, self.base_size * 16 // factor)
        
        self.up1 = Up(self.base_size * 16, self.base_size * 8 // factor, bilinear)
        self.up2 = Up(self.base_size * 8, self.base_size * 4 // factor, bilinear)
        self.up3 = Up(self.base_size * 4, self.base_size * 2 // factor, bilinear)
        self.up4 = Up(self.base_size * 2, self.base_size, bilinear)
        self.outc = OutConv(self.base_size, n_classes)

    def forward(self, x):
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        
        logits = self.outc(x)
        return logits




""" dmt/transforms/harmonizer.py 
Contains the ImageHarmonizer class which takes in various image data types & 
standardizes them to a Sample. This way, images of different types can be
processed, and returned as their original type.
"""


import warnings
import copy
import torch
import numpy as np
import nibabel as nib
import SimpleITK as sitk

from dmt.data.image_base import Image
from dmt.data.images import VectorImage2D, ScalarImage2D, ScalarImage3D
from dmt.data.samples.sample import Sample
from dmt.utils.parse import parse_bool


class ImageHarmonizer:
    """ Handles the conversion of various Image types to be stored in a 
    universal format (i.e. the Sample) for easy processing. Also supports
    outputting of the same data-type when processing is done. This is a core
    component of transforms. """
    
    input_types = ('tensor', 'array', 'sitk', 'image', 'sample')
    
    def __init__(
            self, 
            data, 
            include_keys=None, 
            image_key='main_image',
            copy=True
            ):
        """
        Args:
            data: Can be a tensor, array, sitk image, Image object, or Sample.
            include_keys: list of keys to check size compatibility & copy
                data over if input data is a dict or Sample.
            image_key: key-name to be used in sample.
                Only used if input data is not a sample. 
            copy: Flag to copy data components when getting a new sample.
                If harmonizer.sample is called n times, n copies are returned.
        """
        
        self.data = data
        self.image_key = image_key  # key to store Image in new sample
        self.copy = parse_bool(copy, 'copy')
        
        self.is_dict = False
        self.is_tensor = False
        self.is_array = False
        self.is_sitk = False
        self.is_image = False
        self.is_sample = False
        
        self._sample = self._get_sample()
        self.include_keys = self._parse_include_keys(self._sample, include_keys)
        
    @property
    def sample(self):
        """ Gets the sample (the universal image data type). """
        if not self.copy:
            return self._sample
        return copy.copy(self._sample)
    
    @property
    def consistent_shape(self):
        """ Returns the spatial size of the images. """
        sizes = set()
        for k in self.include_keys:
            sizes.add(tuple(self._sample[k].shape))
        assert sizes  # sanity
        if len(sizes) > 1:
            return None
        return np.array(sizes.pop())
    
    def _get_sample(self):
        data = self.data
        if isinstance(data, dict) and not isinstance(data, Sample):
            sample_d = {k: self._parse_data(v) for k, v in data.items()}
            sample = Sample(sample_d)
            self.is_dict = True
        elif isinstance(data, np.ndarray):
            image = self._parse_data(data)
            sample = Sample({self.image_key: image})
            self.is_array = True
        elif isinstance(data, torch.Tensor):
            image = self._parse_data(data)
            sample = Sample({self.image_key: image})
            self.is_tensor = True
        elif isinstance(data, sitk.SimpleITK.Image):
            image = self._parse_data(data)
            sample = Sample({self.image_key: image})
            self.is_sitk = True
        elif isinstance(data, Image):
            sample = Sample({self.image_key: data})
            self.is_image = True
        elif isinstance(data, Sample):
            sample = data
            self.is_sample = True
        else:
            msg = ('Only tensors, ndarrays, sitk images, Image objects, '
                   'Samples, or dictionaries of the previous are valid. '
                   f'You gave {type(data)}') 
            raise ValueError(msg)
        
        assert sample.num_images > 0, 'You did not give a single image.'
        assert sum([self.is_dict, self.is_tensor, self.is_array, self.is_sitk,
                    self.is_image, self.is_sample]) == 1  # sanity
        return sample
    
    def get_output(self, transformed_sample):
        """ Return data in same data-type as given in init. """
        msg = '"transformed_sample" must be a Sample.'
        assert isinstance(transformed_sample, Sample), msg
        
        if self.is_dict:
            out = {}
            for k in self.data.keys():
                out[k] = transformed_sample[k]
            return out
        elif self.is_array or self.is_tensor:
            image_type = infer_image_type(self.data)
            image = transformed_sample[self.image_key]
            out_type = self.data.dtype
            cf = True if image_type.is_channel_first else False
            if self.is_array:
                return image.get_array(image.image, out_type, channel_first=cf)
            return image.get_tensor(image.image, out_type, channel_first=cf)
        elif self.is_sitk:
            image = transformed_sample[self.image_key]
            return image.image
        elif self.is_image:
            return transformed_sample[self.image_key]
        elif self.is_sample:
            return transformed_sample
        raise RuntimeError('Something is very wrong')
    
    def _parse_data(self, data):
        """ Create Image object from generic image data types. """
        inferrable_types = (np.ndarray, torch.Tensor, sitk.SimpleITK.Image)
        if isinstance(self.data, inferrable_types):
            image_type = infer_image_type(data)
            if image_type.is_2d:
                if image_type.is_vector:
                    return VectorImage2D(data)
                return ScalarImage2D(data)
            return ScalarImage3D(data)
        elif isinstance(self.data, Image):
            return Image
        else:
            return data  # dict or Sample entry is not an Image
    
    def _parse_include_keys(self, sample, include_keys):
        if self.is_dict or self.is_sample:
            image_keys = sample.get_image_keys(include_labels=True)
            if not include_keys:
                return image_keys
            final_include_keys = []
            for k in include_keys:
                if k not in image_keys:
                    msg = (f'Given key ({k}) in "include_keys" is either not '
                           f'in the input data or is not an Image')
                    warnings.warn(msg)
                    continue
                final_include_keys.append(k)
            msg = ('The keys in "include_keys" you gave are all useless. '
                   'Either none of them were in data or none were images.')
            assert final_include_keys, msg
            return final_include_keys
        else:
            return [self.image_key]


def infer_image_type(image):
    """ Through some basic assumptions, try to guess whether an image is 
    (1) 2d or 3d, (2) vector or scalar, or (3) channel first / last
    """
    from collections import namedtuple
    type_args = ['is_2d', 'is_vector', 'is_channel_first']
    ImageType = namedtuple('ImageType', type_args)
    
    if isinstance(image, (torch.Tensor, np.ndarray)):
        # using shape to infer image type
        shape = [d for d in tuple(image.shape) if d != 1]  # squeezed shape
        assert len(shape) in (2, 3), 'Only 2D or 3D images supported'
        if len(shape) == 2:
            is_2d, is_vector, is_channel_first = True, False, None
        else:
            if 3 in shape:
                if shape[0] == 3:
                    is_2d, is_vector, is_channel_first = True, True, True
                    assert shape[-1] != 3, f'Cannot infer shape {shape}'
                elif shape[-1] == 3:
                    is_2d, is_vector, is_channel_first = True, True, False
                    assert shape[0] != 3, f'Cannot infer shape {shape}'
                else:  # assume it is a 3D image with 3 as a middle dim
                    is_2d, is_vector, is_channel_first = False, False, None
            else:
                is_2d, is_vector, is_channel_first = False, False, None
    elif isinstance(image, Image):
        if isinstance(image, ScalarImage3D):
            is_2d = False
            is_vector = False
        else:
            assert isinstance(image, (ScalarImage2D, VectorImage2D))
            is_2d = True
            is_vector = True if isinstance(image, VectorImage2D) else False
        is_channel_first = None
    elif isinstance(image, sitk.SimpleITK.Image):
        dims = image.GetDimension()
        assert dims in (2, 3), 'Only 2D or 3D images supported.'
        is_2d = True if dims == 2 else False
        pixel_depth = image.GetNumberOfComponentsPerPixel()
        assert pixel_depth in (1, 3), 'Wtf kind of image is this'
        is_vector = True if pixel_depth > 1 else False
        is_channel_first = None
    elif isinstance(image, nib.Nifti1Image):
        is_2d = False
        is_vector = False
        is_channel_first = None  # N/A
    else:
        msg = ('Only tensors, ndarrays, Image objects, sitk image, nib images'
               f'types are supported for inferring. You gave {type(image)}') 
        raise ValueError(msg)
    
    image_type = ImageType(is_2d, is_vector, is_channel_first)
    return image_type




import torch
import numpy as np
from .samplers_base import RandomSampler, PatchSamplerScalar3D


class UniformSampler(RandomSampler):
    
    def get_probability_map(self, harmonizer):
        return np.ones(*harmonizer.consistent_shape)
    
    def _generate_patches(self, harmonizer, num_patches=None):
        image_size = harmonizer.consistent_shape
        lower_range = image_size - self.patch_size
        patches_left = num_patches if num_patches is not None else True
        while patches_left:
            crop_lower_indices = [torch.randint(v + 1, (1,)).item() 
                                  for v in lower_range]
            lower_indices_array = np.array(crop_lower_indices)
            yield self.crop(harmonizer, lower_indices_array, self.patch_size)
            if num_patches is not None:
                patches_left -= 1
                
                
class UniformSampler3D(UniformSampler, PatchSamplerScalar3D):
    pass


""" dmt/data/loading/patch_loader.py

A patch-loader similar to the one in torchio except the queue is continuously
being filled (as opposed to only being filled when it's empty). 
"""

import os
import psutil
import time
import logging, warnings
import torch
import torch.multiprocessing as torch_mp
import multiprocessing as python_mp

from dmt.utils.parallel import is_process_running
from dmt.utils.parse import parse_bool, parse_int, parse_nonnegative_int
from dmt.data.loading.collate import default_collate_fn


def example_loading_worker(example_loader, example_queue, shuffle):
    
    examples_storage = set()
    def put_example(ex):
        """ Only called when shuffling is required. """
        if ex is None:
            example_queue.put(examples_storage.pop())
        else:
            examples_storage.add(ex)
            if len(examples_storage) >= 10:
                example_queue.put(examples_storage.pop())
    
    pid = os.getpid()
    ppid = os.getppid()
    this_process = psutil.Process(pid)
    
    start = time.time()
    for i, examples in enumerate(example_loader):
        # print(f'Putting ex {i+1} in Q ({time.time() - start:.2f} sec).')
        # mem_usage = this_process.memory_info()[0]/2.**30
        # logging.debug(f'OTMWorker: Using {mem_usage:.2f} GB in PID {pid}')
        
        if not shuffle:
            for ex in examples:
                example_queue.put(ex)
        else:
            for ex in examples:
                put_example(ex)
        start = time.time()
    
    while examples_storage:
        put_example(None)
    logging.info(f'\n💀 OTMLoader-Side: done loading examples. Exiting now!')


class OneToManyLoader:
    """ Multiprocessing batch loader which creates multiple batch examples from
        a single dataset sample (multiple as in at least one). """
    
    def __init__(
            self,
            dataset,
            sample_processing_fn,  # contains sampler & crop transforms
            examples_per_sample,
            example_collate_fn=default_collate_fn,
            batch_size=16,
            shuffle_samples=True,
            shuffle_patches=True,
            num_workers=3,  # num of workers to load volumes & do transforms
            headstart=False,  # start loading immediately after loader init
            example_queue_maxsize=128,
            drop_last=True
            ):
        """
        Args:
            dataset: Torch-Dataset-like object that collects & preprocesses
                sample data.
            sampler: Samples one or multiple patches given sample data.
            batch_size: number of sampler outputs to collate into a batch.
                Note if sampler outputs 3 patches, batch size = 2 would
                result in a batch with 6 tensors.
            shuffle_samples: Flag to shuffle samples during loading.
            shuffle_patches: Flag to shuffle patches before collation.
            num_workers: Number of processes used for data loading.
            headstart: Flag to start workers right away.
        """
        
        if python_mp.get_start_method() != 'fork':
            msg = ('With multiprocess dataloading, you should use the "fork" '
                   'start method for faster worker initialization times and ' 
                   'reduced memory usage (assuming you are only reading from '
                   'dataset components & not modifying them).')
            warnings.warn(msg)
        
        # Multi-processing to load images faster
        self.dataset = dataset
        self.num_workers = parse_nonnegative_int(num_workers, 'num_workers')
        self.shuffle_samples = parse_bool(shuffle_samples, 'shuffle_samples')
        
        # Sample processing and example loading
        self.sample_loader = torch.utils.data.DataLoader(
            dataset, batch_size=1, num_workers=self.num_workers,
            collate_fn=sample_processing_fn, shuffle=self.shuffle_samples
        )
        self.examples_per_sample = parse_nonnegative_int(examples_per_sample,
                                                         'examples_per_sample')
        self.batch_size = parse_nonnegative_int(batch_size, 'batch_size')
        
        # Example sampling and batch collation
        self.shuffle_patches = parse_bool(shuffle_patches, 'shuffle_patches')
        self.drop_last = parse_bool(drop_last, 'drop_last')
        self.example_collate_fn = example_collate_fn
        self.example_queue_maxsize = parse_int(example_queue_maxsize, 
                                               'example_queue_maxsize')
        
        maxsize = self.example_queue_maxsize
        self._example_queue = torch.multiprocessing.Queue(maxsize=maxsize)
        self._example_loader = None
        self._batch_count = None
        self._loading = False
        self._need_queue_clear = False
        self.iter_runs = 0
        
        if headstart:
            self.start_loading()
    
    @property
    def num_samples(self):
        return len(self.dataset)
    
    @property
    def num_batches(self):
        N_examples = self.num_samples * self.examples_per_sample
        N_full_batches = N_examples // self.batch_size
        if self.drop_last:
            return N_full_batches
        mod = N_examples % self.batch_size
        return N_full_batches + bool(mod)
    
    @property
    def examples_per_epoch(self):
        if self.drop_last:
            return self.num_batches * self.batch_size
        return self.num_samples * self.examples_per_sample
    
    @property
    def loading(self):
        return self._loading
    
    @property
    def is_queue_empty(self):
        return self._example_queue.empty()
    
    @property
    def is_queue_full(self):
        return self._example_queue.full()
    
    @property
    def worker_pid(self):
        if self._example_loader is not None:
            return self._example_loader.pid
        return None
    
    @property
    def worker_memory(self):
        if self._example_worker is not None:
            process = psutil.Process(self.worker_pid)
            mem_usage_mb = process.memory_info().rss / (1024 ** 2)
            return mem_usage_mb
        return None
        
    @property
    def main_pid(self):
        return os.getpid()
    
    def __len__(self):
        return self.num_batches
    
    ### ------ #     Main API     # ----- ###
    
    def __iter__(self):
        self.batch_count = 0
        self.start_loading()
        return self
    
    def __next__(self):
        if self.batch_count >= self.num_batches:
            self.stop_loading()
            raise StopIteration
        else:
            is_last_batch = self.batch_count == self.num_batches - 1
            if is_last_batch and not self.drop_last:
                batch_size = self.examples_per_epoch % self.batch_size
            else:
                batch_size = self.batch_size
            
            start = time.time()
            examples = []
            for i in range(batch_size):
                example = self._example_queue.get(block=True)
                examples.append(example)
            # print(f'  OTM-Iter: {time.time() - start:.2f}s to get examples.')
            
            start = time.time()
            batch = self.example_collate_fn(examples)  
            # print(f'  OTM-Iter: {time.time() - start:.2f}s to collate batch.')
            self.batch_count += 1
            return batch
    
    
    ### ------ #   Continuous Patch Loading & Multiprocessing  # ----- ###
    
    def start_loading(self):
        not_first_iter = self.iter_runs > 0
        if self.loading and self._example_loader is None:
            raise RuntimeError(f'Loader should be set when loading flag is.')
        
        if self.loading:
            assert self._example_loader is not None
            return
        
        if self._need_queue_clear:
            self._clear_queue(self._example_queue)
            self._need_queue_clear = False
            
        args = (self.sample_loader, self._example_queue, self.shuffle_patches)
        target = example_loading_worker
        self._example_loader = torch_mp.Process(target=target, args=args)
        self._example_loader.start()
        self._loading = True
        logging.info(f'🏗️ OTMLoader: successfully started loading process.')
    
    def stop_loading(self):
        self._loading = False
        if self._example_loader is None:
            return
        if self._example_loader.is_alive():
            self._example_loader.kill()
            self._example_loader.join()
            assert not self._example_loader.is_alive()
            logging.info(f'🏗️ OTMLoader: successfully killed loader process.')
        
        self._example_loader = None
        self.iter_runs += 1
        self._need_queue_clear = True
    
    def _clear_queue(self, Q):
        # clear queue
        from queue import Empty
        try:
            disposal_count = 0
            while True:
                Q.get(block=False)
                disposal_count += 1
                logging.info(f'🏗️ OTMLoader-Main: Clearing item from queue.. '
                             f'Cleared {disposal_count} items.')
        except Exception as e:
            if isinstance(e, Empty):
                logging.info('OTOLoader: Successfully cleared batch queue.')
                logging.info('🏗️ OTMLoader: Successfully cleared batch queue.')
            elif not self.batch_queue.empty():
                raise e
        
        




""" Module dmt/data/samples/image_base.py

Contains the base class for all images (2D/3D Scalar/Vector).

Design Notes
 - WARNING: treat dot notation as getting items only (primary implementation
   of kwargs is in a dict); setting items with dot will not update dict entries!
 - Two input categories
    1. file (str_path) <- serialization depends on permanent_load argument
    2. image (sitk, numpy, tensor) <- serialized by default
 - Image properties are lazy loaded (loaded if available on input, if it 
    requires reading then Image attributes are populated when accessed) 
        Properties: shape, spacing, origin, direction, affine.
- For data access, here is what you need to know for function overloading:
    1. self.read_image(): path -> sitk_im  (only if path is inputed)
    2. self.get_sitk(): sitk_im -> sitk_im (serves are processing of sitk)
    3. self.get_array/tensor(): sitk_im -> array/tensor 

"""

from __future__ import annotations
from abc import ABC, abstractmethod

import os, sys
import logging
import weakref
import pathlib
import pprint
from numpy.lib.type_check import _asfarray_dispatcher

import torch
import PIL
import numpy as np
import SimpleITK as sitk

from ..constants import (
    IMTYPE_2_SCALAR_SITK, 
    IMTYPE_2_VECTOR_SITK, 
    IMTYPE_2_NP, 
    IMTYPE_2_TORCH)
from ..utils.io import images2d as im2d_utils, images3d as im3d_utils
from ..utils.images import visualize as viz_utils


# Module-specific Constants
VALID_INPUT_TYPES = (str, pathlib.Path, sitk.SimpleITK.Image, 
                     np.ndarray, torch.Tensor)


class Image(dict):
    """ Base class for an image (2D/3D scalar/vector, including masks). 
    
    Under the hood, image backend uses SimpleITK, has near universal support. 
    
    Args:
        path_or_data: str_path, np, sitk, or torch_tensor
            Tensor assumed to be channel_last (like standard numpy).
        container_type: sitk type to store path_or_data as 
            This is a very important setting that informs whether color channels
            are needed or not during data retrieval of any type. Also, informs
            the 'is_vector' property.
        permanent_load: if true, saves sitk image in self.image for fast loading
            Only in effect if a path is given through path_or_data. 
        **kwargs: additional image properties to be added to object
    """
    
    
    # Class Variables
    _id_counter = 0  # global runtime ID generator for all Image subclasses
    _gid_to_instance = weakref.WeakValueDictionary()
    _reserved_attributes = ['gid', 'path', 'given_type', 'container_type', 
                        'shape', 'spacing', 'origin', 'direction', 'affine',
                        'name', 'pixelid', 'is_2d', 'is_vector', 
                        'sitk_type', 'type', 'memory', 'sitk_image']
    
    def __init__(
            self,
            path_or_data,  # already verified in subclass __init__
            container_type,  # sitk type to store as (in: data) or read (in: path)
            permanent_load=False,  # only applies if input is a path
            **kwargs
            ):
        
        # Global Image ID tracking
        self.gid = Image._id_counter
        Image._gid_to_instance[self.gid] = self
        Image._id_counter += 1
        
        # Load sitk image backend and default attributes
        self.given_type = path_or_data.__class__
        self.container_type = container_type
        self.permanent_load = permanent_load  # updated in _process_input
        
        # Lazy-loaded image attributes to be updated in parse_input
        self._sitk_image = None  # either None (if path given) or sitk image
        self.path = None  # this is None unless updated or given path
        self._shape = None  # lazy-loaded unless data is given as input
        self._spacing = None  # " "  Same for all attributes here & below
        self._origin = None
        self._direction = None
        self._affine = None
        self._process_input(path_or_data, container_type, permanent_load)        
        
        super().__init__(**kwargs)  # update Image attributes to dict
        self.update_attributes()  # make dot notation available
    
    
    ### ------ #  Data Access + Custom Functionality  # ----- ###
    
    def read_image(self, path, sitk_type):
        """ OVERLOAD for custom loading. This is the default implementation.
        Loads an image from file into an sitk image.
            NOTE: this is only called to get sitk image if path is given.
            If any other data was given, then self.sikt_image will be used.
        Args:
            path: str or pathlib to image file
            sitk_type: sitk type to be casted as & to fill isVector argument
        Returns:
            sitk_image
        """
        sitk_type = self._parse_sitk_type(sitk_type)
        path = self._parse_path(path)
        
        sitk_image = sitk.ReadImage(path, sitk_type)
        return sitk_image
    
    def get_array(self, sitk_im, out_type, channel_first=False):
        """ OVERLOAD this default functionality to get array from stik.
        This is probably the most important function in data retrieval. 
            - Given the sitk either from input, file, array, or tensor,
                this function is in charge of preprocessing it & getting array.
            - get_tensor() depends on this
        
        Args:
            sitk_im: SimpleITK.Image which is the default structure in which
                images are stored or read from file.
            out_type: numpy dtype to cast output of conversion to.
            channel_first: If it's a vector image, return channel first format?
        """
        msg = f'Given type "{out_type}" is not a valid numpy dtype.'
        assert out_type in IMTYPE_2_NP.values(), msg
        
        ### Preprocess sitk (do resampling, etc. here)
        if sitk_im.GetPixelID() != self.container_type:
            sitk_im = sitk.Cast(sitk_im, self.container_type)
        
        ### Get numpy array
        array = sitk.GetArrayFromImage(sitk_im)
        if array.dtype != out_type:
            array = array.astype(out_type)
        
        if self.is_vector and channel_first:
            array = np.moveaxis(array, -1, 0)
        return array
        
    def get_tensor(self, sitk_im, out_type, channel_first=True):
        """ OVERLOAD this default functionality to get tensor from np. """
        msg = f'Given type "{out_type}" is not a valid torch tensor dtype.'
        assert out_type in IMTYPE_2_TORCH.values(), msg
        
        general_type_key = self.type
        array_type = IMTYPE_2_NP[general_type_key]
        array = self.get_array(sitk_im, array_type)
        tensor = torch.from_numpy(array)
        
        # Channel-first if is vector
        if self.is_vector and channel_first:
            first_dims = tuple(range(tensor.ndim - 1))
            tensor = tensor.permute(-1, *first_dims)
        tensor = tensor.to(out_type)
        return tensor
    
    def save(self, file):
        """ OVERLOAD if you want saving functionality. 
        You could make use of self.array or self.sitk_image to get image data. 
        """
        sitk.WriteImage(self.sitk_image, file)
        
    def plot(self):
        """ OVERLOAD if you want plotting functionality. 
        You could make use of self.array to get image data. 
        """
        raise NotImplementedError()
    
    
    ### ------ #  Image Properties / Default Attributes  # ----- ###
    
    @property
    def sitk_image(self):
        """ Gets raw SimpleITK.Image object. If this object is None when a path
        is given, then an sitk object is read and returned without interally
        saving.
        """
        sitk_im = self._sitk_image
        if self._sitk_image is None:
            sitk_im = self.read_image(self.path, self.container_type)
            if self._affine is None:  # take the chance to update
                self._update_image_attributes(sitk_im=sitk_im) 
        return sitk_im
    
    @property
    def array(self):
        """ Gets numpy array from output type dervied from give storage type."""
        sitk_im = self.sitk_image
        if self.sitk_image is None:
            sitk_im = self.read_image(self.path, self.container_type)
            if self._affine is None:  # take the chance to update
                self._update_image_attributes(sitk_im=sitk_im) 
        
        # Get corresponding np dtype from sitk type
        general_type_key = self.type
        array_type = IMTYPE_2_NP[general_type_key]
        array = self.get_array(sitk_im, array_type)
        return array
    
    @property
    def tensor(self):
        sitk_im = self.sitk_image
        if self.sitk_image is None:
            sitk_im = self.read_image(self.path, self.container_type)
            if self._affine is None:  # take the chance to update
                self._update_image_attributes(sitk_im=sitk_im) 
            
        # Get corresponding tensor dtype from sitk type
        general_type_key = self.type
        tensor_type = IMTYPE_2_TORCH[general_type_key]
        tensor = self.get_tensor(sitk_im, tensor_type)
        return tensor
    
    
    ### ------ # Image Properties # ----- ###
    
    @property
    def shape(self):
        if self._shape is None:
            self._update_image_attributes()
        return self._shape
    
    @property
    def spacing(self):
        if self._spacing is None:
            self._update_image_attributes()
        return self._spacing
    
    @property
    def origin(self):
        if self._origin is None:
            self._update_image_attributes()
        return self._origin
    
    @origin.setter
    def origin(self, value):
        old_origin = self.origin

        msg = '"Image.origin" must be set to a flattened np.ndarray.'
        assert isinstance(value, np.ndarray), msg
        
        msg = (f'Given "origin" shape {value.shape} does not match with the '
               f'old "origin" shape {old_origin.shape}.')
        assert old_origin.shape == value.shape, msg
        
        self._origin = value
    
    @property
    def direction(self):
        if self._direction is None:
            self._update_image_attributes()
        return self._direction
    
    @property
    def affine(self):
        if self._affine is None:
            self._update_image_attributes()
        return self._affine
    
    @affine.setter
    def affine(self, value):
        old_affine = self.affine
        
        msg = 'Image.affine must be set to a flattened np.ndarray.'
        assert isinstance(value, np.ndarray), msg
        
        msg = (f'Given "affine" shape {value.shape} does not match with the '
               f'old affine shape {old_affine.shape}.')
        assert old_affine.shape == value.shape, msg
        
        self._affine = value
    
    @property
    def memory(self): 
        """ Get memory use in bytes. """
        tot_bytes = 0
        for o in self.__dict__.values():
            if isinstance(o, sitk.SimpleITK.Image):
                tot_bytes += Image.get_sitk_memory(o)
            else:
                tot_bytes += sys.getsizeof(o)
        return tot_bytes
    
    @property
    def type(self):
        if self.is_vector:
            sitk_type_dict = IMTYPE_2_VECTOR_SITK 
        else:
            sitk_type_dict = IMTYPE_2_SCALAR_SITK
        type_key = None
        for k, v in sitk_type_dict.items():
            if v == self.container_type:
                type_key = k
                break
        assert type_key is not None, 'Oops, check implementation.'
        return type_key
    
    @property
    def sitk_type(self):
        return sitk.GetPixelIDValueAsString(self.container_type)
    
    @property
    def is_vector(self):
        """ Returns True if given type is a vector image, False otherwise. """
        sitk_type = self.container_type
        isVector = True if sitk_type in IMTYPE_2_VECTOR_SITK.values() else False
        return isVector
    
    @property
    def is_2d(self):
        return self._is_2d
    
    @property
    def pixelid(self):
        return self.container_type
    
    @property
    def name(self):
        path = str(pathlib.Path(self.path).name) if self.path else ''
        name = f"{self.__class__.__name__}_gid{self.gid}"
        if path:
            name += f'_{path}'
        return name
    
    def _update_image_attributes(self, sitk_im=None):
        # print(f"COSTLY!! UPDATING ATTRIBUTES!")
        if sitk_im is None:
            sitk_im = self.sitk_image
        self._shape = np.array(tuple(sitk_im.GetSize()))
        self._spacing = np.array(tuple(sitk_im.GetSpacing()))
        self._origin = np.array(tuple(sitk_im.GetOrigin()))
        self._direction = np.array(tuple(sitk_im.GetDirection()))
        self._affine = Image.get_affine(sitk_im)
    
    def _process_input(self, path_or_data, container_type, permanent_load):
        r""" Processes input and sets image attributes if available. 
        Args:
            path_or_data: str_path, sitk_image, np_array, or torch_tensor
            container_type: casted to this sitk type if input is path
            permanent_load: boolean that is only considered if input is a path
        Returns: None
        """
        if isinstance(path_or_data, (str, pathlib.Path)):
            path_or_data = str(path_or_data)  # convert to str if pathlib
            
            msg = ('If data input is a path, then container_type must be '
                   'given as an sitk image type.')
            assert container_type is not None, msg
            
            container_type = self._parse_sitk_type(container_type)
            path_or_data = self._parse_path(path_or_data)
            
            self.path = path_or_data
            if permanent_load:
                self._sitk_image = self.read_image(path_or_data, container_type)
        elif isinstance(path_or_data, sitk.SimpleITK.Image):
            self._sitk_image = path_or_data  # type will be cast below
        elif isinstance(path_or_data, torch.Tensor):
            path_or_data = path_or_data.detach().cpu().numpy()  # handled later
        
        #  NOTE: PILs are error prone when converting so do it manually.
        # elif isinstance(path_or_data, PIL.Image.Image):
        #     path_or_data = np.array(path_or_data)
        
        if isinstance(path_or_data, np.ndarray):
            isVector = self.is_vector
            self._sitk_image = sitk.GetImageFromArray(path_or_data, 
                                                      isVector=isVector)
        
        # Update image attributes if self.image is defined
        if self._sitk_image is not None:
            if self._sitk_image.GetPixelID() != container_type:
                self._sitk_image = sitk.Cast(self._sitk_image, container_type)
            self.permanent_load = True
            self._update_image_attributes()
        
        # Sanity Checks
        if permanent_load:
            msg = 'Img not there'
            assert isinstance(self._sitk_image, sitk.SimpleITK.Image), msg
        if self._sitk_image is None:
            assert self.path, 'If image is None, then path must be set'
            assert os.path.isfile(self.path), 'File {self.path} does not exist.'
            
    def _parse_sitk_type(self, sitk_type):
        valid_sitk_types = list(IMTYPE_2_SCALAR_SITK.values()) + \
                           list(IMTYPE_2_VECTOR_SITK.values())
        msg = f'Given sitk type {str(sitk_type)} isn\'t valid.'
        assert sitk_type in valid_sitk_types, msg
        return sitk_type
    
    def _parse_path(self, path):
        msg = 'Path must be a string or pathlib.Path'
        assert isinstance(path, (str, pathlib.Path)), msg
        
        path = str(path)
        assert os.path.isfile(path), f'File does not exist: {path}'
        return path
    
    
    @staticmethod
    def get_affine(sitk_im):
        ndim = sitk_im.GetDimension()
        affine = np.eye(ndim + 1)

        rotation = np.array(sitk_im.GetDirection()).reshape(ndim, ndim)
        origin = np.array(sitk_im.GetOrigin())
        spacing = np.array(sitk_im.GetSpacing())
        flip_xy = np.diag([1] * ndim)  # used to switch between LPS and RAS
        # print(rotation, origin, spacing, flip_xy)

        M = np.dot(flip_xy, rotation) * spacing
        T = np.dot(flip_xy, origin)
        affine = np.eye(ndim + 1)
        affine[:ndim, :ndim] = M
        affine[:ndim, ndim] = T
        return affine
    
    @staticmethod
    def get_sitk_memory(sitk_im):
        """ Get total memory used by a sitk image object in bytes. """
        import re
        regex = r"Size: [\d]+"
        matches = re.findall(regex, str(sitk_im))[0]
        return int(matches.split()[-1])
        
    @staticmethod
    def get_image_from_id(img_id):
        assert isinstance(img_id, int), f"Image ID must be an integer!"
        if img_id < 0 or img_id >= Image._id_counter:
            return None
        return Image._gid_to_instance[img_id]
    
    def __repr__(self):
        path = '' 
        if self.path:
            path = f"w/ image: {str(pathlib.Path(self.path).name)}"
        sitk_type = sitk.GetPixelIDValueAsString(self.container_type)
        
        affine = self._affine
        if affine is not None:
            affine = affine.flatten()
        
        attrs = {k: v for k, v in self.items() 
                    if k not in self._reserved_attributes and k[0] != '_'}
        attrs_str = pprint.pformat(attrs, indent=1, compact=True)
        
        string = (
            f"{self.__class__.__name__} (gid={self.gid}) {path}\n"
            f"  Given Format (type={self.given_type}) \n"
            f"  Type: {sitk_type}, Shape {self._shape}, Spacing {self._spacing} \n"
            f"  Origin {self._origin}, Direction {self._direction} \n"
            f"  Affine {affine} \n"
            f"  Tot Mem = {self.memory:,} bytes. \n"
            f"Other Attributes: \n {attrs_str} \n"
        )
        return string
    
    ### Functions below offers compatible/correct dict functionality
    ###   e.g. must update __dict__ after every update & override default funcs
    
    def __setattr__(self, name, value):
        """ Setting attrs in dot notation requires updating dict. """
        name = self._parse_new_attribute(name)
        self[name] = value
        super().__setattr__(name, value)
    
    def __getitem__(self, item):
        item = super().__getitem__(item)
        gid = None if not hasattr(self, 'gid') else self.gid
        logging.debug(f'Sample {gid}: Getting item: {item}')
        return item
    
    def __setitem__(self, key, value):
        key = self._parse_new_attribute(key)
        super().__setitem__(key, value)
        self.update_attributes()
        gid = None if not hasattr(self, 'gid') else self.gid
        logging.debug(f'Sample {gid}: Getting item: {value}')
        
    def update(self, *args, **kwargs):
        if args:
            if len(args) > 1:
                msg = f"Update expected at most 1 arguments, got {len(args)}."
                raise TypeError(msg)
            for k, v in dict(args[0]).items():
                k = self._parse_new_attribute(k)
                self[k] = v
        for k, v in kwargs.items():
            k = self._parse_new_attribute(k)
            self[k] = v
        self.update_attributes()
            
    def setdefault(self, key, value=None):
        if key not in self:
            self[key] = value
        self.update_attributes()
        return self[key]
    
    def update_attributes(self):
        # Allows attribute access through dot notation, e.g. image.spacing
        self.__dict__.update(self)
        
    def _parse_new_attribute(self, attribute):
        return attribute
        
        # TODO gatekeep user attribute setting
        msg = (f'Given attribute "{attribute}" is a reserved attribute '
               f'for class {type(self).__name__}.')
        assert attribute not in self._reserved_attributes, msg
        
        msg = f'Given attribute "{attribute}" cannot start with "_".'
        assert attribute[0] != '_', msg
        
        return attribute
    




from .base_model import BaseModel

from .unets import unet2d, unet3d, bigunet2d, denseunet2d


__all__ = ['BaseModel']


model_names_2d = {
    'unet2d',
    'denseunet2d',
    'bigunet2d',
}

model_names_3d = {
    'unet3d', 
}


def get_model(name, **kwargs):
    
    if name in model_names_2d:
        
        if name == 'unet2d':
            model = unet2d.get_model(**kwargs)
        elif name == 'denseunet2d':
            model = denseunet2d.get_model(**kwargs)
        elif name == 'bigunet2d':
            model = bigunet2d.get_model(**kwargs)
            
        
    elif name in model_names_3d:
        
        if name == 'unet3d':
            model = unet3d.get_model(**kwargs)
        
    else:
        msg = (f'"{name}" is not a valid model name. \n2D models include: '
               f'{model_names_2d}. \n3D models include: {model_names_3d}.')
        raise ValueError(msg)
    
    return model
        
    # name = cfg['model']['name']
    # settings = cfg['model'][name]

    # ds_name = cfg['data']['name']
    # out_channels = len(cfg['data'][ds_name]['classnames'])
    # in_channels = 0
    # for entry in cfg['train']['transforms']:
    #     if entry[0] == 'normmeanstd':
    #         in_channels = len(entry[1][0])
    # assert in_channels > 0, f"Could not get input dim from normmeanstd transform"

    print(f"  > Fetching {name} model..", end='')
    if name == 'unet':
        bilinear = settings['bilinear']
        model = unet.UNet(in_channels, out_channels, bilinear=bilinear)
    elif 'bigunet' in name:
        bilinear = settings['bilinear']
        base_size = settings['base_size']
        model = bigunet.BigUNet(in_channels, out_channels, bilinear=bilinear,
                    base_size=base_size)
    elif name == 'nestedunet':
        model = nestedunet.NestedUNet(
            out_channels, 
            input_channels=in_channels,
            deep_supervision=cfg['model']['nestedunet']['deepsup']
        )
    elif name == 'r2attentionunet':
        model = r2attentionunet.R2Attention_UNet(
            in_ch=in_channels, out_ch=out_channels, t=2
        )
    elif name == 'segdensenet':
        model = segdensenet.get_model(
            str(settings['layers']), 
            num_classes=out_channels, 
            pretrained=settings['pretrained']
        )
    elif name == 'denseunet':
        model = denseunet.get_model(
            str(settings['layers']),
            pretrained=settings['pretrained'],
            num_classes=out_channels, 
            deconv=settings['deconv']
        )
    elif name == 'msradenseunet':
        model = msradenseunet.get_model(
            str(settings['layers']),
            pretrained=settings['pretrained'],
            num_classes=out_channels, 
            deconv=settings['deconv']
        )
    elif name == 'imsradenseunet':
        model = imsradenseunet.get_model(
            str(settings['layers']),
            pretrained=settings['pretrained'],
            num_classes=out_channels, 
            deconv=settings['deconv']
        )
    # elif name == 'denseunet_deepsup':
    #     model = denseunet_deepsup.get_model(
    #         str(settings['layers']),
    #         pretrained=settings['pretrained'],
    #         num_classes=out_channels, 
    #         deconv=settings['deconv']
    #     )
    else:
        

    print(f"\t{name} initialized ({model.param_counts[0]} params).")
    return model




import subprocess


def get_gpu_memory_map():
    """Get the current gpu usage.

    Returns
    -------
    usage: dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    """
    result = subprocess.check_output(
        [
            'nvidia-smi', '--query-gpu=memory.used',
            '--format=csv,nounits,noheader'
        ], encoding='utf-8')
    # Convert lines into a dictionary
    gpu_memory = [int(x) for x in result.strip().split('\n')]
    gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    
    return gpu_memory_map



import warnings
import logging
import textwrap


OUT = set(('print', 'warning', 'log_debug', 'log_info', 'log_warning',
           'log_error', 'log_critical'))
# OUT_STREAM = {
#     'print': print,
#     'warning': warnings.warn,
#     'log_debug': logging.debug,
#     'log_info': logging.info,
#     'log_warning': logging.warning,
#     'log_error': logging.error,
#     'log_critical': logging.critical,
# }


def header_one(message, width=80, out='print'):
    divider = '▓' * width
    if len(message) > width - 4:
        center_message = wrap_and_center(message, width=width-4)
    else:
        center_message = ('☯ ' + message + ' ☯').center(width, ' ')
        # center_message = '⛩️⛩️⛩️ ' + message + ' ⛩️⛩️⛩️'
    log('\n\n' + divider + '\n' + center_message + '\n' + divider + '\n\n', out)
    

def header_two(message, width=80, out='print'): 
    divider = '━' * width
    if len(message) > width - 4:
        center_message = wrap_and_center(message, width=width-4)
    else:
        center_message = ('● ' + message + ' ●').center(width, ' ')
        # center_message = '⛩️⛩️⛩️ ' + message + ' ⛩️⛩️⛩️'
    log('\n\n' + divider + '\n' + center_message + '\n' + divider + '\n\n', out)
    
def header_three(message, width=80, out='print'):
    if len(message) > width - 6:
        messages = textwrap.fill(message, width - 6).split('\n')
        for i, msg in enumerate(messages):
            if i == 0 or i == len(messages) - 1:
                messages[i] = ('○ ' + messages[i] + ' ○').center(width, '⋯')
            else:
                messages[i] = messages[i].center(width, ' ')
        center_message = '\n'.join(messages)
    else:
        center_message = ('○ ' + message + ' ○').center(width, '⋯')
    log('\n' + center_message + '\n', out)

def subsection(message, width=80, out='print'):
    message = '✼ ' + message + ' ✼'
    underline = '┈' * min(width, len(message))
    message = textwrap.fill(message, width=width)
    message = message.replace('\n', '\n   ')
    log('\n' + message + '\n' + underline, out)
    
def subsubsection(message, width=80, out='print'):
    message = textwrap.fill(message, width=width)
    message = message.replace('\n', '\n   ')
    log('\n-[' + message + ']-\n', out)

def wrap_and_center(string, width=80):
    strings = textwrap.wrap(string, width)
    strings = [string.center(width, ' ') for string in strings]
    return '\n'.join(strings)


def log(message, out):
    out = parse_out(out)
    # OUT_STREAM[out](message)
    if out == 'print':
        print(message)
    elif out == 'warning':
        warnings.warn(message)
    elif out == 'log_debug':
        logging.debug(message)
    elif out == 'log_info':
        logging.info(message)
    elif out == 'log_warning':
        logging.warning(message)
    elif out == 'log_error':
        logging.error(message)
    elif out == 'log_critical':
        logging.critical(message)


def parse_out(out):
    out = out.lower()
    msg = f'Given out "{out}" is not valid. Choices: {OUT}.'
    assert out.lower() in OUT, msg
    return out


    
if __name__ == '__main__':
    strings = ('Step 1: Collect Training Components'*4, 
               'Step 1: Collect Training Components'*12, 
               'Step 2: Training'
    )
    
    for sep in (header_one, header_two, header_three, subsection, subsubsection):
        for string in strings:
            sep(string)



import numbers
import matplotlib.pyplot as plt
import numpy as np
import collections
import wandb


LOWER_IS_BETTER = ['loss', 'fn', 'fp', 'hausdorff', 'distance']

class WandBTracker:
    """ All-in-one concise experiment tracking after each epoch completion.
        - Tracks all essential iteration and epoch metrics.
        - Automatically logs epoch stats to wandb.
        - If wand_settings not given, acts as a local tracker.
    Parameters
        wandb_settings = {'name': .., 'project': .., 'config':.., 'notes':.. }
        metric_names = ['list', 'of', 'metric', 'names']
    """

    def __init__(self, metric_names=[], wandb_settings={}):
        self.metrics_d = { k: [] for k in metric_names } if metric_names else {}
        if wandb_settings:
            self.use_wandb = True
            self._setup_wandb(wandb_settings)
        else:
            self.use_wandb = False
    
    def update(self, metrics_d, log=False, summarize={}, verbose=False,
               after_epoch=0, force_summarize=False):
        summary_keys = []
        for k, item in metrics_d.items():
            if k not in self.metrics_d:
                if verbose:
                    print(f" (StatTracker) Adding stat: {k}, Val: {item}")
                self.metrics_d[k] = [item]
            else:
                self.metrics_d[k].append(item)
                if verbose:
                    print(f"(StatTracker) Adding {k}: {item}")
                    
            if self.use_wandb and summarize and k in summarize['triggers']:
                summary_keys.append(k)
                
        for k in summary_keys:
            max_better = False if sum([m.lower() in k for m in \
                LOWER_IS_BETTER]) else True
            if force_summarize or \
              self.is_best(k, max_better=max_better, after_epoch=after_epoch):
                stepnum = len(self.metrics_d[k]) - 1
                trigger = k

                update_d = {}
                for kk in summarize['saves']:
                    if kk not in self.metrics_d:
                        print(f" (WandDBTracker) WARNING: summary met {kk} not valid.")
                        continue
                    name = f"trigger({trigger})_step{stepnum}({kk})"
                    update_d[name] = self.metrics_d[kk][-1]
                wandb.run.summary.update(update_d)
            
#         if self.use_wandb and sum([s in k for s in summarize]):
#             max_better = False if sum([m.lower() in k for m in \
#                 LOWER_IS_BETTER]) else True
#             if self.best(k, max_better=max_better) == item:
#                 stepnum = len(self.metrics_d[k]) - 1
#                 wandb.run.summary.update({
#                     'best_' + k: item, 'best_' + k + '_step': stepnum})
        
        if log and self.use_wandb:
            wandb.log(metrics_d)
    
    def best(self, metric_name, max_better=True):
        if metric_name not in self.metrics_d:
            print(f" (WandBTracker) Given k({metric_name}) not valid.")
            return None
        metric_hist = self.metrics_d[metric_name]
        return max(metric_hist) if max_better else min(metric_hist)

    def is_best(self, metric_name, max_better=True, after_epoch=0):
        if metric_name not in self.metrics_d:
            print(f" (WandBTracker) Given k({metric_name}) not valid.")
            return False
        metric_hist = self.metrics_d[metric_name]
        if len(metric_hist) <= after_epoch:
            return False
        metric_hist = metric_hist[after_epoch:]
        best = max(metric_hist) if max_better else min(metric_hist)
        return True if metric_hist[-1] == best else False

    def best_by_metrics(self, metric_name, max_better=True, after_epoch=0):
        if metric_name not in self.metrics_d:
            print(f" (WandBTracker) Given k({metric_name}) not valid.")
            return None
        N_epochs = len(self.metrics_d[metric_name])
        met_ep = [(self.metrics_d[metric_name][i], i) for i in range(N_epochs)]
        reverse = True if max_better else False
        met_ep = sorted(met_ep, key=lambda x: x[0], reverse=reverse)
        
        best_ep = -1
        for m, e in met_ep:
            if e > after_epoch:
                best_ep = e
                break
        if best_ep == -1:
            print(f" (WandBTracker) no metrics fit the criteria given.")
            return None

        best_mets = {}
        best_mets['epoch'] = best_ep
        for k, mets in self.metrics_d.items():
            if len(mets) > best_ep:
                best_mets[k] = mets[best_ep]
        return best_mets

    def plot(self, keys=None):
        names, vals = [], []
        if keys:
            for k in keys:
                if k not in self.metrics_d:
                    continue
                names.append(k)
                vals.append(self.metrics_d[k])
        else:
            for k, v in self.metrics_d.items():
                if v and not isinstance(v[0], collections.Sequence):
                    names.append(k)
                    vals.append(vals)

        num_cols = 4
        num_rows = len(names) // num_cols + 1
        fig = plt.figure(figsize=(20, 5*num_rows))
        for i, name in enumerate(names):
            ax = fig.add_subplot(num_rows, num_cols, i+1)
            ax.set_title(name)
            ax.plot(vals[i])
        plt.show()

    def resume(self, wandb_settings={}):
        """ Called when restarting after training session(s) has/have finished."""
        if not self.use_wandb or not hasattr(self, 'wandb_run_id'):
            raise ValueError(f"Prev run did not have a wandb session.")
        if wandb_settings:
            wandb.init(
                resume=self.wandb_run_id,
                project=wandb_settings['project'],
                name=wandb_settings['name'], 
                config=wandb_settings['config'],
                notes=None if 'notes' not in wandb_settings else wandb_settings['notes']
            )
        else:
            wandb.init(resume=self.wandb_run_id)
            
    def _setup_wandb(self, wandb_settings):
        print(" > Initializing Weights and Biases run..")
        cfg = wandb_settings['config']
        name = cfg['experiment']['id'] + '_' + cfg['experiment']['name']
        self.wandb_run_id = wandb.util.generate_id()
        print("\t- ID: {self.wandb_run_id}  - Name: {name}")
        wandb.init(
            id=self.wandb_run_id,
            resume='allow',
            project=wandb_settings['project'],
            name=wandb_settings['name'], 
            config=wandb_settings['config'],
            notes=None if 'notes' not in wandb_settings else wandb_settings['notes']
        )


class EpochMeters:
    """ Updates every iteration and keeps track of accumulated stats. """
    
    def __init__(self):
        self.accums = {}
        self.ns = {}

    def update(self, metrics_d, n=1):
        for k, item in metrics_d.items():
            if k not in self.accums:
                self.accums[k] = item
                self.ns[k] = n
                continue
            self.accums[k] += item
            self.ns[k] += n

    def avg(self, no_avg=[]):
        ret = {}
        for k, v in self.accums.items():
            if k in no_avg:
                ret[k] = v
            else:
                ret[k] = v/self.ns[k]
        return ret
        

# Basic Tests
if __name__ == '__main__':
    import sys

    metrics = {
        'train_loss': [.98, .76, .56, .43, .58],
        'test_loss': [.99, .77, .25, .44, .59],
        'overall_fp': [300, 250, 80, 100, 120],
        'overall_dice': [.98, .67, .68, .69, .69]
    }
    cfg = {'experiment': {'id': 'test0000', 'name': 'hello'}}
    tracker = WandBTracker(wandb_settings={
        'name': '0003_lr02', 'project': 'bibm2020', 'id': '0000', 'config': cfg
    })
    for epoch in range(len(metrics['train_loss'])):
        epmeter = EpochMeters()
        for it in range(1, 5):
            epmeter.update({
                'TPs': np.array([1,2,3]),
                'F1': 1.1,
                'F1s': np.array([.2,.4,.6])
            })
        tracker.update({k: v[epoch] for k, v in metrics.items()},
            log=True, summarize=['_loss', 'overall_dice'])
    import IPython; IPython.embed(); import sys; sys.exit(1)



""" Module classification.py (By: Charley Zhang, Nov 2020)
Collection of multi-class, multi-label, and reconstruction losses for 
general image classification tasks.

Resources:
https://discuss.pytorch.org/t/soft-cross-entropy-loss-tf-has-it-does-pytorch-have-it/69501
"""

import torch, torch.nn as nn
import torch.nn.functional as F
import numpy as np


__all__ = ['HardCrossEntropy', 'SoftCrossEntropy', 'SoftmaxMSE']



class FocalLoss(nn.Module):
    def __init__(self, alpha=0.25, gamma=2):
        super(FocalLoss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        print(f" (FocalLoss) Initiated with alpha {alpha} & gamma {gamma}.")

    def forward(self, output, target):
        '''Focal loss.
        Args:
          output: (tensor) sized [N,D].
          target: (tensor) sized [N,].
        Return:
          (tensor) focal loss.
        '''
        alpha = self.alpha
        gamma = self.gamma
        x = output
        
        to_one_hot = torch.eye(output.shape[1])
        t = to_one_hot[target.data.cpu()].to(output.device)

        p = x.sigmoid()
        pt = p*t + (1-p)*(1-t)         # pt = p if t > 0 else 1-p
        w = alpha*t + (1-alpha)*(1-t)  # w = alpha if t > 0 else 1-alpha
        w = (w * (1-pt).pow(gamma)).detach()
        loss = F.binary_cross_entropy_with_logits(x, t, weight=w, reduction='sum')
        return loss

    
class ClassBalancedLoss(nn.Module):
    def __init__(self, samples_per_class, lossname='focal', beta=0.999, gamma=1):
        """ https://github.com/vandit15/Class-balanced-loss-pytorch/
        Parameters
          samples_per_class: A python list of size [no_of_classes].
          lossname: string. One of "sigmoid", "focal", "softmax".
          beta: float. Hyperparameter for Class balanced loss.
          gamma: float. Hyperparameter for Focal loss.
        """
        assert lossname in ('sigmoid', 'focal', 'softmax')
        
        super(ClassBalancedLoss, self).__init__()
        self.samples_per_class = samples_per_class
        self.lossname = lossname
        self.beta = beta
        self.gamma = gamma
        
        print(f" (ClassBalancedLoss) Initiated with class_counts={samples_per_class}\n"
              f"\t loss={lossname}, beta={beta}, gamma={gamma}.")
    
    def forward(self, logits, labels):
        """Compute the Class Balanced Loss between `logits` and the ground truth `labels`.
        Class Balanced Loss: ((1-beta)/(1-beta^n))*Loss(labels, logits)
        where Loss is one of the standard losses used for Neural Networks.
        Parameters
            logits: A float tensor of size [batch, no_of_classes].
            labels: A int tensor of size [batch].
        Returns:
            cb_loss: A float tensor representing class balanced loss
        """
        loss_type = self.lossname
        beta = self.beta
        gamma =self.gamma
        
        no_of_classes = logits.shape[1]
        samples_per_cls = self.samples_per_class
        
        # Original Implementation
        effective_num = 1.0 - np.power(beta, samples_per_cls)
        weights = (1.0 - beta) / np.array(effective_num)
        weights = weights / np.sum(weights) * no_of_classes

        labels_one_hot = F.one_hot(labels, no_of_classes).float().to(logits.device)

        weights = torch.tensor(weights).float().to(logits.device)
        weights = weights.unsqueeze(0)
        weights = weights.repeat(labels_one_hot.shape[0],1) * labels_one_hot
        weights = weights.sum(1)
        weights = weights.unsqueeze(1)
        weights = weights.repeat(1,no_of_classes)

        if loss_type == "focal":
            cb_loss = self.focal_loss(labels_one_hot, logits, weights, gamma)
        elif loss_type == "sigmoid":
            cb_loss = F.binary_cross_entropy_with_logits(
                input = logits,target = labels_one_hot, 
                weights = weights
            )
        elif loss_type == "softmax":
            pred = logits.softmax(dim = 1)
            cb_loss = F.binary_cross_entropy(
                input = pred, target = labels_one_hot, 
                weight = weights
            )
        return cb_loss
    
    def focal_loss(self, labels, logits, alpha, gamma):
        """Compute the focal loss between `logits` and the ground truth `labels`.
        Focal loss = -alpha_t * (1-pt)^gamma * log(pt)
        where pt is the probability of being classified to the true class.
        pt = p (if true class), otherwise pt = 1 - p. p = sigmoid(logit).
        Args:
          labels: A float tensor of size [batch, num_classes].
          logits: A float tensor of size [batch, num_classes].
          alpha: A float tensor of size [batch_size]
            specifying per-example weight for balanced cross entropy.
          gamma: A float scalar modulating loss from hard and easy examples.
        Returns:
          focal_loss: A float32 scalar representing normalized total loss.
        """    
        BCLoss = F.binary_cross_entropy_with_logits(
            input = logits, target = labels,
            reduction = "none"
        )

        if gamma == 0.0:
            modulator = 1.0
        else:
            modulator = torch.exp(-gamma * labels * logits - gamma * torch.log(1 + 
                torch.exp(-1.0 * logits)))

        loss = modulator * BCLoss

        weighted_loss = alpha * loss
        focal_loss = torch.sum(weighted_loss)

        focal_loss /= torch.sum(labels)
        return focal_loss



class HardCrossEntropy(nn.Module):
	""" Default nn.CrossEntropy with hard integer labels.
	Expects input & target to be BxC (batchsize x num_classes).
	"""

	def __init__(self, weights=None, reduction='mean', ignore_index=-100):
		assert reduction in ['none', 'mean', 'sum'], f"{reduction} not valid."
		
		super(HardCrossEntropy, self).__init__()
		self.reduction = reduction
		self.weights = torch.tensor(weights).float() if weights is not None else None
		self.ce = nn.CrossEntropyLoss(weight=self.weights, reduction=reduction,
			ignore_index=ignore_index) 
		print(f"\t (HardCE) Initialized w/{reduction} reduction & weights: {weights}")

	def forward(self, output, target):
		"""
		Parameters:
			output (BxC float tensor) - logits or probs for each batch example
			target (B long tensor) - class index 
		"""
		if self.weights is not None:
			assert self.weights.shape[0] == output.shape[1]
		assert output.shape[0] == target.shape[0], (output.shape, target.shape)
		
		# if logits:
		# 	output = output.softmax(-1)
		# assert output.sum().abs().item() - output.shape[0] < 10**-5, \
		# 	(output.sum(), output.shape[0])

		loss = self.ce(output, target.long())
		return loss


class SoftCrossEntropy(nn.Module):
	""" Real statistical CrossEntropy that deals with 2 batched probability
	distributions. 
	"""

	def __init__(self, reduction='mean', weights=None):
		assert reduction in ['none', 'mean', 'sum'], f"{reduction} not valid."
		
		super(SoftCrossEntropy, self).__init__()
		self.reduction = reduction
		self.weights = torch.tensor(weights) if weights is not None else None
		print(f"\t (SoftCE) Initialized w/{reduction} reduction & weights: {weights}")

	def forward(self, output, target):
		"""
		Parameters:
			output (BxC float tensor) - model output logits
			target (BxC float tensor) - target distribution
		"""
		N, C = output.shape
		if self.weights is not None:
			assert self.weights.shape[0] == output.shape[1] == target.shape[1]

		assert output.sum().abs().item() - N < 10**-5
		assert target.sum().abs().item() - N < 10**-5

		logprobs = F.log_softmax(output, dim=1)
		cum_loss = -(target * logprobs)

		if self.reduction == 'none':
			return cum_loss
		elif self.reduction == 'sum':
			return cum_loss.sum()
		else:
			return cum_loss.sum() / N


class MSE(nn.Module):
	
	def __init__(self, reduction='mean'):
		assert reduction in ['none', 'mean', 'sum'], f"{reduction} not valid."

		super(MSE, self).__init__()
		self.mse = nn.MSELoss(reduction=reduction)
		print(f"\t (MSE) Initialized w/{reduction} reduction.")

	def forward(self, output, target):
		loss = self.mse(output, target)
		return loss


class SoftmaxMSE(nn.Module):
	
	def __init__(self, weights=None, reduction='mean'):
		assert reduction in ['none', 'mean', 'sum'], f"{reduction} not valid."
		
		super(SoftmaxMSE, self).__init__()
		self.reduction = reduction
		self.weights = torch.tensor(weights) if weights is not None else None
		print(f"\t (SoftmaxMSE) Initialized w/{reduction} reduction & weights: {weights}")

	def forward(self, output, target):
		"""
		Parameters:
			output (BxC float tensor) - network output logits
			target (BxC float tensor) - target logits
		"""
		assert output.shape == target.shape, (output.shape, target.shape)
		N, C = target.shape
		output_softmax = F.softmax(output, dim=1)
		target_softmax = F.softmax(target, dim=1)

		loss = (output_softmax - target_softmax) ** 2
		if self.weights is not None:
			loss = loss * self.weights.to(loss.device)
		
		if self.reduction == 'none':
			return loss
		elif self.reduction == 'sum':
			return loss.sum()
		else:
			return loss.sum() / (target.shape[0] * target.shape[1])
		
		# Experimental Code.. (not run)
		mse = F.mse_loss(output_softmax, target_softmax, reduction='none').sum(1) / C
		ent_mask = (- output_softmax * torch.log(output_softmax)).sum(1) > 0.1
		out = mse * ent_mask.float()

		return out.sum() / ent_mask.count_nonzero()
		



def relation_mse_loss(activations, ema_activations):
    """Takes softmax on both sides and returns MSE loss
    Note:
    - Returns the sum over all examples. Divide by the batch size afterwards
      if you want the mean.
    - Sends gradients to inputs but not the targets.
    """

    assert activations.size() == ema_activations.size()

    activations = torch.reshape(activations, (activations.shape[0], -1))
    ema_activations = torch.reshape(ema_activations, (ema_activations.shape[0], -1))

    similarity = activations.mm(activations.t())
    norm = torch.reshape(torch.norm(similarity, 2, 1), (-1, 1))
    norm_similarity = similarity / norm

    ema_similarity = ema_activations.mm(ema_activations.t())
    ema_norm = torch.reshape(torch.norm(ema_similarity, 2, 1), (-1, 1))
    ema_norm_similarity = ema_similarity / ema_norm

    similarity_mse_loss = (norm_similarity-ema_norm_similarity)**2
    return similarity_mse_loss

